# Interactive apps to show solutions of Riemann problems.

# Left and right states in the phase plane can be dragged
# and dropped, as well as the time bar on the x-t plane.
# MPLD3: D3-javascript plugin for matplotlib http://mpld3.github.io/
# Questions contact Mauricio Del Razo S. @maojrs: maojrs@gmail.com
"""
In order to run the following code in python, try running the tests at 
the end of the code. If you are using the jupyter notebooks, remember 
switching mpld3.show() to mpld3.display()
"""


import numpy as np
import jinja2
import json
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.path as mpath
import matplotlib.patches as mpatches

import mpld3
from mpld3 import plugins, utils

# Shallow water interactive phase plane plugin in mpld3
class PPlaneNLPlugin(plugins.PluginBase):
    JAVASCRIPT = r"""
    // init custom PPLaneNL plugin
    mpld3.register_plugin("drag", PPlaneNLPlugin);
    PPlaneNLPlugin.prototype = Object.create(mpld3.Plugin.prototype);
    PPlaneNLPlugin.prototype.constructor = PPlaneNLPlugin;
    PPlaneNLPlugin.prototype.requiredProps = ["id", "idmpoint", 
                                              "idg", "iditer", "iditer_charac", "idtime", "idhmax", "idoffm",
                                              "idhugol", "idhugor", "idintcl", "idintcr",
                                              "idhugol2", "idhugor2", "idintcl2", "idintcr2", 
                                              "idqlm", "idqmm", "idqrm", 
                                              "idshock1", "idshock2", "idrar1", "idrar2",
                                              "idrar1a", "idrar2a", "idrar1b", "idrar2b",
                                              "idrar1c", "idrar2c", "idrar1d", "idrar2d",
                                              "idtimedot", "idtimeline",
                                              "idq1", "idq2"];
    PPlaneNLPlugin.prototype.defaultProps = {}
    function PPlaneNLPlugin(fig, props){
        mpld3.Plugin.call(this, fig, props);
        mpld3.insert_css("#" + fig.figid + " path.dragging",
                         {"fill-opacity": "1.0 !important",
                          "stroke-opacity": "1.0 !important"});
    };

    // Call draw function, this function is being looped all the time
    PPlaneNLPlugin.prototype.draw = function(){
        // Get elements into script variables
        var obj = mpld3.get_element(this.props.id);
        var midpoint = mpld3.get_element(this.props.idmpoint);
        var g = this.props.idg;
        var iter = this.props.iditer;
        var iter_charac = this.props.iditer_charac;
        var time = this.props.idtime;
        var hmax = this.props.idhmax;
        var offm = this.props.idoffm;
        var hugol = mpld3.get_element(this.props.idhugol);
        var hugor = mpld3.get_element(this.props.idhugor);
        var intcl = mpld3.get_element(this.props.idintcl);
        var intcr = mpld3.get_element(this.props.idintcr);
        var hugol2 = mpld3.get_element(this.props.idhugol2);
        var hugor2 = mpld3.get_element(this.props.idhugor2);
        var intcl2 = mpld3.get_element(this.props.idintcl2);
        var intcr2 = mpld3.get_element(this.props.idintcr2);
        var qlm = mpld3.get_element(this.props.idqlm);
        var qmm = mpld3.get_element(this.props.idqmm);
        var qrm = mpld3.get_element(this.props.idqrm);
        var shock1 = mpld3.get_element(this.props.idshock1);
        var shock2 = mpld3.get_element(this.props.idshock2);
        var rar1 = mpld3.get_element(this.props.idrar1);
        var rar2 = mpld3.get_element(this.props.idrar2);
        var rar1a = mpld3.get_element(this.props.idrar1a);
        var rar2a = mpld3.get_element(this.props.idrar2a);
        var rar1b = mpld3.get_element(this.props.idrar1b);
        var rar2b = mpld3.get_element(this.props.idrar2b);
        var rar1c = mpld3.get_element(this.props.idrar1c);
        var rar2c = mpld3.get_element(this.props.idrar2c);
        var rar1d = mpld3.get_element(this.props.idrar1d);
        var rar2d = mpld3.get_element(this.props.idrar2d);
        var timedot = mpld3.get_element(this.props.idtimedot);
        var timeline = mpld3.get_element(this.props.idtimeline);
        var q1 = mpld3.get_element(this.props.idq1);
        var q2 = mpld3.get_element(this.props.idq2);
        
        // Set initial conditions for javascript calculations
        //var qleft = obj.offsets[0];
        //var qright = obj.offsets[1];
        //var qmid = midpoint.offsets[0];
        var offx = obj.ax.x(offm);
        var offy = offx;
        // Define initial values of hl, hr, hul and hur
        var hl = obj.offsets[0][0];
        var hul = obj.offsets[0][1];
        var hr = obj.offsets[1][0];
        var hur = obj.offsets[1][1];
        hmfinal = 0;
        humfinal = 0;
        
        // Main d3 drag function
        var drag = d3.behavior.drag()
            .origin(function(d) { return {x:obj.ax.x(d[0]),
                                          y:obj.ax.y(d[1])}; }) 
            .on("dragstart", dragstarted)
            .on("drag", dragged)
            .on("dragend", dragended);
        
        // Dragtime function for draggable time-dot (analog to main)
        var dragtime = d3.behavior.drag()
            .origin(function(d) { return {x:timedot.ax.x(d[0]),
                                          y:timedot.ax.y(d[1])}; })                              
            .on("dragstart", dragstarted)
            .on("drag", dragged)
            .on("dragend", dragended);
    
        // Set elements of ql and qr draggable points and call main drag function
        obj.elements()
           .data(obj.offsets)
           .style("cursor", "default")
           .call(drag);
            
        // Set elements for timedot draggable point and call dragtime function  
        timedot.elements()
           .data(timedot.offsets)
           .style("cursor", "default")
           .call(dragtime);

        // Begin phi and phi prime function
        function phi(h,hl,hr,hul,hur) {
            var ul = hul/hl;
            var ur = hur/hr;
            if (h < hl) {
                var termIC1 = 2.0*Math.sqrt(g*hl);
                var termIC2 = 2.0*Math.sqrt(g*h);
                var phil = ul + termIC1 - termIC2;
                var philp = -Math.sqrt(g/h);
            }
            if (h >= hl) {
                var termHL1 = (h - hl);
                var termHL2 = Math.sqrt(0.5*g*(1.0/h + 1.0/hl));
                var termHL3 = termHL1*termHL2;
                var phil = ul - termHL3 ;
                var nume = termHL2*4*h*h;
                var philp = -termHL2 + termHL1*g/nume;
            }
            if (h < hr) {
                var termIC1 = 2.0*Math.sqrt(g*hr);
                var termIC2 = 2.0*Math.sqrt(g*h);
                var phir = ur - termIC1 + termIC2;
                var phirp = Math.sqrt(g/h);
            }
            if (h >= hr) {
                var termHL1 = (h - hr);
                var termHL2 = Math.sqrt(0.5*g*(1.0/h + 1.0/hr));
                var termHL3 = termHL1*termHL2;
                var phir = ur + termHL3 ;
                var nume = termHL2*4*h*h;
                var phirp = termHL2 - termHL1*g/nume;
            }
            var phi = phil-phir;
            var phiprime = philp - phirp
            var um = 0.5*(phil + phir);
            return [phi,phiprime,um];
        }
        
        // Begin Newton iteration
        function newton(hstar,hl,hr,hul,hur) {
            var hn = hstar;
            var error = 1;
            while (error > 0.005) {
                var hold = hn;
                var phin = phi(hn,hl,hr,hul,hur);
                var hn = hold - phin[0]/phin[1];
                var um = phin[2];
                error = Math.abs(hold - hn);
            }
            return [hn,um];
        }
        
        // Calculate hugoniot loci
        function hugoloci(hm,hside,huside,sign) {
                var termHL1 = hm*(hm - hside);
                if (hm == 0) {
                    var termHL2 = 1.0; }
                else {
                   var termHL2 = Math.sqrt(0.5*g*(1.0/hm + 1.0/hside));}
                   var termHL3 = termHL1*termHL2;
                   var hloci = hm*huside/hside + sign*termHL3;
            return hloci;} 
            
        // Calculate integral curve
        function integralcurve(hm,hside,huside,sign) {
                var termIC1 = 2.0*hm*Math.sqrt(g*hside);
                var termIC2 = 2.0*hm*Math.sqrt(g*hm);
                var intcurve = hm*huside/hside - sign*(termIC1 - termIC2);
            return intcurve;}
            
        // Calculate solution plot of h and hu as function of x
        function solplot(hl,hm,hr,hul,hum,hur) {
            var lam = [];
            lam.push([-1000000, 0,0])
            var ul = hul/hl;
            var um = hum/hm;
            var ur = hur/hr;
            var shock1 = false;
            var shock2 = false;
            var qsol1 = d3.range(iter);
            var qsol2 = d3.range(iter);
            if (hm >= hl) {
                var s1 = time*(hum - hul)/(hm - hl);
                shock1 = true;
            } else {
                var s1b = [0,0];
                s1b[0] = time*(ul - Math.sqrt(g*hl));
                s1b[1] = time*(um - Math.sqrt(g*hm));
                s1b.sort(d3.ascending);
            }
            if (hm >= hr) {
                var s2 = time*(hur - hum)/(hr - hm);
                shock2 = true;
            } else {
                var s2b = [0,0];
                s2b[0] = time*(um + Math.sqrt(g*hm));
                s2b[1] = time*(ur + Math.sqrt(g*hr));
                s2b.sort(d3.ascending);
            }
            
            for (var ii=0; ii <2*iter; ii++){
                var xx = q1.data[ii][0];
                // Calculate plot solution for 1 charactersitic (shock or rarefaction)
                if (shock1) {
                    if (xx <= s1){ qsol1[ii] = hl; qsol2[ii] = hul;}
                    else { qsol1[ii] = hm; qsol2[ii] = hum;}
                }
                else {
                    if (xx <= s1b[0]) { qsol1[ii] = hl; qsol2[ii] = hul;}
                    else if (xx <= s1b[1]) {
                        var m1 = (hm - hl)/(s1b[1] - s1b[0]);
                        var m2 = (hum - hul)/(s1b[1] - s1b[0]);
                        qsol1[ii] = m1*(xx - s1b[0]) + hl;
                        qsol2[ii] = m2*(xx - s1b[0]) + hul;}
                    else { qsol1[ii] = hm; qsol2[ii] = hum;}
                }
                // Calculate plot solution for 2 charactersitic (shock or rarefaction)
                if (shock2) {
                    if (xx > s2) {qsol1[ii] = hr; qsol2[ii] = hur;}
                }
                else{
                    if (xx > s2b[0] && xx < s2b[1]) {
                        var m1 = (hr - hm)/(s2b[1] - s2b[0]);
                        var m2 = (hur - hum)/(s2b[1] - s2b[0]);
                        qsol1[ii] = m1*(xx - s2b[0]) + hm;
                        qsol2[ii] = m2*(xx - s2b[0]) + hum;}
                    else if (xx > s2b[1]) { qsol1[ii] = hr; qsol2[ii] = hur;}
                }
            }
            var solution = [qsol1,qsol2];
            return solution;
        }

        // Function to update middle state   
        function update_midstate(){
            var hstar = 0.05*(hl + hr);
            var solution = newton(hstar,hl,hr,hul,hur);
            hmfinal = solution[0];
            humfinal = solution[1]*hmfinal;
            var xx = obj.ax.x(hmfinal);
            var yy = obj.ax.y(humfinal);
            // Update middle state point and marker position
            midpoint.elements().transition().duration(5)
            .attr("transform", "translate(" + [xx, yy] + ")");
            // Move marker
            qmm.elements().transition().duration(1)
              .attr("transform", "translate(" + [xx + 0.7*offx, yy + 0.7*offy] + ")");
        }
        
        // Functon to update xt-plane
        function update_xtplane() {
         // Calculate shcok speeds from R-H conditions
         var lam1 = (humfinal - hul)/(hmfinal - hl);   
         var lam2 = (hur - humfinal)/(hr - hmfinal);   
         var lam1m = humfinal/hmfinal - Math.sqrt(g*hmfinal);
         var lam2m = humfinal/hmfinal + Math.sqrt(g*hmfinal);
         var lam1l = hul/hl - Math.sqrt(g*hl);
         var lam2r = hur/hr + Math.sqrt(g*hr);
         var color1 = "red";
         var color2 = "red";
         var thick1 = 4;
         var thick2 = 4;
         var fan = 0;
         for (var ii=0; ii<iter_charac; ii++) {
             if (hmfinal >= hl) {
                 shock1.data[ii][1] = shock1.data[ii][0]/lam1;
                 rar1.data[ii][1] = rar1.data[ii][0]/lam1;
                 rar1a.data[ii][1] = rar1a.data[ii][0]/lam1;
                 rar1b.data[ii][1] = rar1b.data[ii][0]/lam1;
                 rar1c.data[ii][1] = rar1c.data[ii][0]/lam1;
                 rar1d.data[ii][1] = rar1d.data[ii][0]/lam1;
                 console.log(rar2a.data);
                 color1 = "red";
                 thick1 = 4;
             } else {
                 shock1.data[ii][1] = shock1.data[ii][0]/lam1l;
                 rar1.data[ii][1] = rar1.data[ii][0]/lam1m;
                 fan = Math.abs(lam1m - lam1l)/5;
                 rar1a.data[ii][1] = rar1a.data[ii][0]/(lam1l + fan);
                 rar1b.data[ii][1] = rar1b.data[ii][0]/(lam1l + 2*fan);
                 rar1c.data[ii][1] = rar1c.data[ii][0]/(lam1l + 3*fan);
                 rar1d.data[ii][1] = rar1d.data[ii][0]/(lam1l + 4*fan);
                 color1 = "blue";
                 thick1 = 1;
             }
             if (hmfinal >= hr) {
                 shock2.data[ii][2] = shock2.data[ii][0]/lam2;
                 rar2.data[ii][2] = rar2.data[ii][0]/lam2;
                 rar2a.data[ii][2] = rar2a.data[ii][0]/lam2;
                 rar2b.data[ii][2] = rar2b.data[ii][0]/lam2;
                 rar2c.data[ii][2] = rar2c.data[ii][0]/lam2;
                 rar2d.data[ii][2] = rar2d.data[ii][0]/lam2;
                 color2 = "red";
                 thick2 = 4;
            } else {
                 shock2.data[ii][2] = shock2.data[ii][0]/lam2r;
                 rar2.data[ii][2] = rar2.data[ii][0]/lam2m;
                 fan = Math.abs(lam2r - lam2m)/5;
                 rar2a.data[ii][2] = rar2a.data[ii][0]/(lam2m + fan);
                 rar2b.data[ii][2] = rar2b.data[ii][0]/(lam2m + 2*fan);
                 rar2c.data[ii][2] = rar2c.data[ii][0]/(lam2m + 3*fan);
                 rar2d.data[ii][2] = rar2d.data[ii][0]/(lam2m + 4*fan);
                 color2 = "blue";
                 thick2 = 1;
            }                     
         }
         // Do transitions
         shock1.elements().transition().duration(5)
             .attr("d", shock1.datafunc(shock1.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         shock2.elements().transition().duration(5)
             .attr("d", shock2.datafunc(shock2.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
         rar1.elements().transition().duration(5)
             .attr("d", rar1.datafunc(rar1.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         rar2.elements().transition().duration(5)
             .attr("d", rar2.datafunc(rar2.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
         rar1a.elements().transition().duration(5)
             .attr("d", rar1a.datafunc(rar1a.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         rar2a.elements().transition().duration(5)
             .attr("d", rar2a.datafunc(rar2a.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
         rar1b.elements().transition().duration(5)
             .attr("d", rar1b.datafunc(rar1b.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         rar2b.elements().transition().duration(5)
             .attr("d", rar2b.datafunc(rar2b.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
         rar1c.elements().transition().duration(5)
             .attr("d", rar1c.datafunc(rar1c.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         rar2c.elements().transition().duration(5)
             .attr("d", rar2c.datafunc(rar2c.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
         rar1d.elements().transition().duration(5)
             .attr("d", rar1d.datafunc(rar1d.data))
             .style("stroke", color1)
             .style("stroke-width", thick1);
         rar2d.elements().transition().duration(5)
             .attr("d", rar2d.datafunc(rar2d.data))
             .style("stroke", color2)
             .style("stroke-width", thick2);
        }
        
        // Function to update solution plots
        function update_solplots() {
            var qsol = solplot(hl,hmfinal,hr,hul,humfinal,hur);
                for (var ii=0; ii<2*iter; ii++){
                    q1.data[ii][1] = qsol[0][ii]; 
                    q2.data[ii][2] = qsol[1][ii]; 
                }
                // Do transitions
                 q1.elements().transition().duration(5)
                     .attr("d", q1.datafunc(q1.data));
                 q2.elements().transition().duration(5)
                     .attr("d", q2.datafunc(q2.data));
        }
        
        // Initialize solution with given initial states before interacting
        update_midstate();
        update_xtplane();
        update_solplots();
        
        // Begin drag function
        function dragstarted(d) {
          d3.event.sourceEvent.stopPropagation();
          d3.select(this).classed("dragging", true);
        }
   
        // The drag function called while dragging is happening (meat of code here)
        function dragged(d,i) {
          if (i == 0  || i ==1) {
              // Convert mouse coordinates in drag event (d3.event) to python coordinates d
              d[0] = obj.ax.x.invert(d3.event.x);
              d[1] = obj.ax.y.invert(d3.event.y);             
              // Move ql and qr stored in obj (they have been selected in drag)
              d3.select(this)
                .attr("transform", "translate(" + [d3.event.x,d3.event.y] + ")");
              // If obj corresponds to ql, move all the other left elements 
              }
          if (i==0){
              // Move marker
              qlm.elements().transition().duration(1)
                  .attr("transform", "translate(" + [d3.event.x + offx, d3.event.y + offy] + ")");     
              // Re-calculate inital left variables when dragging 
              hl = obj.offsets[0][0];
              hul = obj.offsets[0][1];
              
              // Draw Hugoniot loci through left state    
              for (var ii=0; ii<iter; ii++) {
                //Left from hl
                hugol.data[ii][0] = ii*d[0]/(1.0*iter);
                var hm = hugol.data[ii][0];
                hugol.data[ii][1] = hugoloci(hm,hl,hul,-1) ;
                //Right from hl
                hugol2.data[ii][0] = d[0] + ii*(hmax-d[0])/(1.0*iter);
                var hm2 = hugol2.data[ii][0];
                hugol2.data[ii][1] = hugoloci(hm2,hl,hul,-1) ;
                }
              // Do transitions      
              hugol.elements().transition().duration(5)
                  .attr("d", hugol.datafunc(hugol.data))
                  .style("stroke-dasharray", ("7,7"));
              hugol2.elements().transition().duration(5)
                  .attr("d", hugol2.datafunc(hugol2.data));
                               
                  
              // Draw integral curve through left state (note index bug for intcl.data[ii][j])
              // and intcl2 it should be j =0,1 not j=1,2. Arrays somehow grow of sixe in mpld3
              for (var ii=0; ii<iter; ii++) {
                //Left from hl
                intcl.data[ii][1] = ii*d[0]/(1.0*iter);
                var hm = intcl.data[ii][1];
                intcl.data[ii][2] = integralcurve(hm,hl,hul,-1);
                //Right from hl
                intcl2.data[ii][1] = d[0] + ii*(hmax-d[0])/(1.0*iter);
                var hm2 = intcl2.data[ii][1];
                intcl2.data[ii][2] = integralcurve(hm2,hl,hul,-1);
              }
              // Do transitions      
              intcl.elements().transition().duration(5)
                  .attr("d", intcl.datafunc(intcl.data));
              intcl2.elements().transition().duration(5)
                  .attr("d", intcl2.datafunc(intcl2.data))
                  .style("stroke-dasharray", ("7,7"));
                  
          }      
          // if element corresponds to qr
          else if (i==1) {
              // Move marker
              qrm.elements().transition().duration(1)
                  .attr("transform", "translate(" + [d3.event.x + offx, d3.event.y + offy] + ")");
              // Re-calculate inital right variables when dragging
              hr = obj.offsets[1][0];
              hur = obj.offsets[1][1];
              
              // Draw Hugoniot loci through right state    
              for (var ii=0; ii<iter; ii++) {
                //Left from hr
                hugor.data[ii][0] = ii*d[0]/(1.0*iter);
                var hm = hugor.data[ii][0];
                hugor.data[ii][1] = hugoloci(hm,hr,hur,1) ;
                //Right from hr
                hugor2.data[ii][0] = d[0] + ii*(hmax-d[0])/(1.0*iter);
                var hm2 = hugor2.data[ii][0];
                hugor2.data[ii][1] = hugoloci(hm2,hr,hur,1) ;
                }
              // Do transitions      
              hugor.elements().transition().duration(5)
                  .attr("d", hugor.datafunc(hugor.data))
                  .style("stroke-dasharray", ("7,7"));
              hugor2.elements().transition().duration(5)
                  .attr("d", hugor2.datafunc(hugor2.data));
                    
              // Draw integral curve through right state (note index bug for intcr.data[ii][j])
              // and intcr2 it should be j =0,1 not j=1,2. Arrays somehow grow of sixe in mpld3   
              for (var ii=0; ii<iter; ii++) {
                //Left from hr
                intcr.data[ii][1] = ii*d[0]/(1.0*iter);
                var hm = intcr.data[ii][1];
                intcr.data[ii][2] = integralcurve(hm,hr,hur,1);
                //Right from hr
                intcr2.data[ii][1] = d[0] + ii*(hmax-d[0])/(1.0*iter);
                var hm2 = intcr2.data[ii][1];
                intcr2.data[ii][2] = integralcurve(hm2,hr,hur,1);
              }
              // Do transitions      
              intcr.elements().transition().duration(5)
                  .attr("d", intcl.datafunc(intcr.data));
              intcr2.elements().transition().duration(5)
                  .attr("d", intcr2.datafunc(intcr2.data))
                  .style("stroke-dasharray", ("7,7"));
            }
            // If time marker is moved
            else if (i==2) {
                // Convert mouse coordinates in drag event (d3.event) to python coordinates d
                d[0] = timedot.ax.x.invert(d3.event.x);
                d[1] = timedot.ax.y.invert(d3.event.y);      
                d3.select(this)
                    .attr("transform", "translate(" + [d3.event.x,d3.event.y] + ")");
                // Calculate timedot position and assign it
                var ty = timedot.ax.y.invert(d3.event.y);
                if (ty >=0) {
                    timeline.data[0][1] = ty;
                    timeline.data[1][1] = ty;
                    timeline.data[2][1] = ty;
                    time = ty;
                }
                // Do transitions
                timeline.elements().transition().duration(5)
                    .attr("d", timeline.datafunc(timeline.data));
            }
            
            // Calculate middle state    
            update_midstate();
                  
            // Calculate solution plots of h and hu
            update_solplots();
            
            // Update characteristic in x-t plane
            update_xtplane();
             
        }
        // End dragging
        function dragended(d) {
          d3.select(this).classed("dragging", false);
        }
    }
    mpld3.register_plugin("drag", PPlaneNLPlugin); 
    """

    def __init__(self, points, midpoint, 
                 g, iters, iter_charac, time, hmax, offm,
                 hugol, hugor, intcl, intcr,
                 hugol2, hugor2, intcl2, intcr2, 
                 qlmarker, qmmarker ,qrmarker, 
                 shock1, shock2, rar1, rar2,
                 rar1a,rar2a,rar1b,rar2b,
                 rar1c,rar2c,rar1d,rar2d,
                 timedot, timeline,
                 q1, q2):
        if isinstance(points, mpl.lines.Line2D):
            suffix = "pts"
        else:
            suffix = None

        self.dict_ = {"type": "drag",
                      "id": utils.get_id(points, suffix),
                      "idmpoint": utils.get_id(midpoint,suffix),
                      "idg" : g,
                      "iditer" : iters,
                      "iditer_charac" : iter_charac,
                      "idtime" : time,
                      "idhmax" : hmax,
                      "idoffm": offm,
                      "idhugol" : utils.get_id(hugol),
                      "idhugor" : utils.get_id(hugor),
                      "idintcl" : utils.get_id(intcl),
                      "idintcr" : utils.get_id(intcr),
                      "idhugol2" : utils.get_id(hugol2),
                      "idhugor2" : utils.get_id(hugor2),
                      "idintcl2" : utils.get_id(intcl2),
                      "idintcr2" : utils.get_id(intcr2),
                      "idqlm": utils.get_id(qlmarker,suffix),
                      "idqmm": utils.get_id(qmmarker,suffix),
                      "idqrm": utils.get_id(qrmarker,suffix),
                      "idshock1" : utils.get_id(shock1),
                      "idshock2" : utils.get_id(shock2),
                      "idrar1" : utils.get_id(rar1),
                      "idrar2" : utils.get_id(rar2),
                      "idrar1a" : utils.get_id(rar1a),
                      "idrar2a" : utils.get_id(rar2a),
                      "idrar1b" : utils.get_id(rar1b),
                      "idrar2b" : utils.get_id(rar2b),
                      "idrar1c" : utils.get_id(rar1c),
                      "idrar2c" : utils.get_id(rar2c),
                      "idrar1d" : utils.get_id(rar1d),
                      "idrar2d" : utils.get_id(rar2d),
                      "idtimedot" : utils.get_id(timedot,suffix),
                      "idtimeline" : utils.get_id(timeline),
                      "idq1" : utils.get_id(q1),
                      "idq2" : utils.get_id(q2)
                      }

# Python part for shallow water interactive part
# Define hugoloci and integral curve functions
def hugoloci(g,hm,hside,huside,sign):
    term1 = hm*(hm - hside)
    hm[hm==0] = hm[hm==0] + 0.00000001
    term2 = np.sqrt(0.5*g*(1.0/hm + 1.0/hside))
    hloci = hm*huside/hside + sign*term1*term2
    return hloci

def intcurve(g,hm,hside,huside,sign):
    term1 = 2.0*hm*np.sqrt(g*hside);
    term2 = 2.0*hm*np.sqrt(g*hm);
    intcurve = hm*huside/hside - sign*(term1 - term2);
    return intcurve

# Define interactive plot routine in python (calls mpld3 plugin)
# Loaded with default values, can be called without any argument
def shallow_water(ql=np.array([3.0, 5.0]), qr=np.array([3.0, -5.0]), g=1.0, time=2.0, tmax=5.0, 
                  hmax=None, humin=None, humax=None):
    # Create figure
    # Create a figure
    #fig, axfull = plt.subplots(2,2, figsize=(13, 12))
    fig, axfull = plt.subplots(2,2, figsize=(10, 8))
    fig.subplots_adjust(left=0.05, right=0.9, bottom=0.1, top=0.9,
                        hspace=0.3, wspace=0.15)

    # Have to do this to avoid issue with mpld3
    ax = axfull[0] # First row of plots
    axsol = axfull[1] # Second row of plost 

    # Calculate dq with ql and qr and other parameters
    dq = np.array([qr[0]-ql[0], qr[1]-ql[1]])
    iters = 100
    iter_charac = 2
    # eps required for bug in mpld3 (ql[0] cannot be the same than qr[0])
    eps = 0.00000001
    
    # Calculate plotting boundaries if not specified
    if hmax is None:
        hmax = max(ql[0],qr[0]) + 7.0
    if humax is None:
        humax = 3*max(abs(ql[1]),abs(qr[1]))
    if humin is None:    
        humin = -3*max(abs(ql[1]),abs(qr[1]))

    # PLOT PHASE PLANE
    xxL = np.linspace(0,ql[0],iters)
    xxL2 = np.linspace(ql[0],hmax,iters)
    xxR = np.linspace(0,qr[0]+eps,iters)
    xxR2 = np.linspace(qr[0]+eps,hmax,iters)

    #Plot midpoint
    qm = -1 +0.0*ql
    midpoint = ax[0].plot(qm[0],qm[1],'ok', alpha=0.9, markersize=8, markeredgewidth=1)

    # Plot hugoloci initial state
    yyL = hugoloci(g,xxL,ql[0],ql[1],-1)
    yyL2 = hugoloci(g,xxL2,ql[0],ql[1],-1)
    yyR = hugoloci(g,xxR,qr[0],qr[1],1)
    yyR2 = hugoloci(g,xxR2,qr[0],qr[1],1)
    hugol = ax[0].plot(xxL,yyL, '--r', linewidth=1.5)
    hugol2 = ax[0].plot(xxL2,yyL2, '-r', linewidth=2, label = 'Hugoniot Loci')
    hugor = ax[0].plot(xxR,yyR, '--r', linewidth=1.5, label = 'Hugoniot Loci (unphysical)')
    hugor2 = ax[0].plot(xxR2,yyR2, '-r', linewidth=2)

    # Plot integral curve initial state
    yyL = intcurve(g,xxL,ql[0],ql[1],-1)
    yyL2 = intcurve(g,xxL2,ql[0],ql[1],-1)
    yyR = intcurve(g,xxR,qr[0],qr[1],1)
    yyR2 = intcurve(g,xxR2,qr[0],qr[1],1)
    intcl = ax[0].plot(xxL,yyL, '-b', linewidth=2, label = 'Integral Curves')
    intcl2 = ax[0].plot(xxL2,yyL2, '--b', linewidth=1.5, label = 'Integral Curves (unphysical)')
    intcr = ax[0].plot(xxR,yyR, '-b', linewidth=2)
    intcr2 = ax[0].plot(xxR2,yyR2, '--b', linewidth=1.5)

    # Plot ql and qr
    points = ax[0].plot([ql[0],qr[0]], [ql[1], qr[1]], 'ok', alpha=0.7, markersize=10, markeredgewidth=1)
    #data = ["q_l", "q_r"] 

    # Plot markers
    offsetx = 0.3*hmax/10
    offsety = -3*offsetx
    qlmarker = ax[0].plot(ql[0]+offsetx, ql[1]+offsety, 'ok', marker=(r"$ q_l $"), markersize=15)
    qmmarker = ax[0].plot(qm[0]+offsetx, qm[1]+offsety, 'ok', marker=(r"$ q_m $"), markersize=15)
    qrmarker = ax[0].plot(qr[0]+offsetx, qr[1]+offsety, 'ok', marker=(r"$ q_r $"), markersize=15)

    # Set axis 1 properties
    ax[0].set_title("Phase plane", fontsize=18)
    ax[0].axis([0,hmax,humin,humax])
    ax[0].set_xlabel('h', fontsize=17)
    ax[0].set_ylabel('hu', fontsize=17)
    #ax[0].set_aspect('equal')
    ax[0].grid(alpha=0.1,color='k', linestyle='--')
    legend = ax[0].legend(loc='upper left', shadow=True, fontsize = 8)


    # PLOT x-t PLANE
    x_xtp = np.linspace(-10,10,iter_charac)
    x_xtp2 = np.linspace(-11,11,iter_charac)
    # Shock speeds lam1 and lam2
    lam2 = ql[1]/ql[0] - np.sqrt(g*ql[0])
    lam1 = qr[1]/qr[0] + np.sqrt(g*qr[0])
    char1 = x_xtp/lam1
    char2 = x_xtp/lam2
    char3 = x_xtp2/lam1
    char4 = x_xtp2/lam2
    shock1 = ax[1].plot(x_xtp, char1, '-k', linewidth=4, label="1 or 2 shock")
    shock2 = ax[1].plot(x_xtp, char2, '-k', linewidth=4)
    rar1 = ax[1].plot(x_xtp2, char3, '-k', linewidth=1, label="1 or 2 rarefaction")
    rar2 = ax[1].plot(x_xtp2, char4, '-k', linewidth=1)
    
    # For other 4 rarefaction lines on each side
    x_xtp3 = np.linspace(-11.1,11.1,iter_charac)
    x_xtp4 = np.linspace(-11.2,11.2,iter_charac)
    x_xtp5 = np.linspace(-11.3,11.3,iter_charac)
    x_xtp6 = np.linspace(-11.4,11.4,iter_charac)
    char1a = x_xtp3/lam1; char2a = x_xtp3/lam2 
    char1b = x_xtp4/lam1; char2b = x_xtp4/lam2 
    char1c = x_xtp5/lam1; char2c = x_xtp5/lam2 
    char1d = x_xtp6/lam1; char2d = x_xtp6/lam2 
    rar1a = ax[1].plot(x_xtp3, char1a, '-k', linewidth=1)
    rar2a = ax[1].plot(x_xtp3, char2a, '-k', linewidth=1)
    rar1b = ax[1].plot(x_xtp4, char1b, '-k', linewidth=1)
    rar2b = ax[1].plot(x_xtp4, char2b, '-k', linewidth=1)
    rar1c = ax[1].plot(x_xtp5, char1c, '-k', linewidth=1)
    rar2c = ax[1].plot(x_xtp5, char2c, '-k', linewidth=1)
    rar1d = ax[1].plot(x_xtp6, char1d, '-k', linewidth=1)
    rar2d = ax[1].plot(x_xtp6, char2d, '-k', linewidth=1)
    
    timedot = ax[1].plot([100000,1000000,9], [-10,-10,time], 'ok' , alpha=0.7, markersize=10)
    timeline = ax[1].plot([-12,0,12], [time, time, time], '--k', linewidth = 3, label="time")

    # Set axis 2 properties
    ax[1].set_title("x-t plane", fontsize=18)
    ax[1].set_xlabel('x', fontsize=17)
    ax[1].set_ylabel('t', fontsize=17)
    ax[1].axis([-10,10,-0,tmax])
    ax[1].grid(alpha=0.1,color='k', linestyle='--')
    legend = ax[1].legend(loc='upper center', shadow=True, fontsize = 8)

    # PLOT SOLUTIONS
    xsol = np.linspace(-10,10,2*iters)
    hsol = 0*xsol + ql[0]
    husol = 0*xsol + ql[1]
    q1 = axsol[0].plot(xsol,hsol, '-k', linewidth = 4, alpha = 1.0)
    q2 = axsol[1].plot(xsol,husol, '-k', linewidth = 4, alpha = 1.0)

    def solplot(xsol,ql,qr,qm,g):
        hl = ql[0];    hm = qm[0];    hr = qr[0]
        ul = ql[1]/hl; um = qm[1]/hm; ur = qr[1]/hr     
        lam = np.empty(4, dtype=float)

    # Set axis 3 properties
    axsol[0].set_title("Depth h at time = t", fontsize=18)
    axsol[0].set_xlabel('x', fontsize=17)
    axsol[0].set_ylabel('h', fontsize=17)
    axsol[0].axis([-10,10,0,hmax])
    axsol[0].grid(alpha=0.1,color='k', linestyle='--')

    # Set axis 4 properties
    axsol[1].set_title("Momentum hu at time = t", fontsize=18)
    axsol[1].set_xlabel('x', fontsize=17)
    axsol[1].set_ylabel('hu', fontsize=17)
    axsol[1].axis([-10,10,humin,humax])
    axsol[1].grid(alpha=0.1,color='k', linestyle='--')


    # Remove defult mpld3 plugins
    plugins.clear(fig)

    # Call mpld3 custom PPLane plugin to interact with plot
    plugins.connect(fig, PPlaneNLPlugin(points[0],midpoint[0], 
                                        g,iters,iter_charac,time, hmax, offsetx,
                                        hugol[0],hugor[0],intcl[0],intcr[0],
                                        hugol2[0],hugor2[0],intcl2[0],intcr2[0],
                                        qlmarker[0],qmmarker[0],qrmarker[0],
                                        shock1[0],shock2[0],rar1[0],rar2[0],
                                        rar1a[0],rar2a[0],rar1b[0],rar2b[0],
                                        rar1c[0],rar2c[0],rar1d[0],rar2d[0],
                                        timedot[0],timeline[0],
                                        q1[0],q2[0]))
    return fig


##########################################

# Plugin for interactive phase plane plot for 2D linear case 
class PPlanePlugin(plugins.PluginBase):
    JAVASCRIPT = r"""
    // init custom PPLane plugin
    mpld3.register_plugin("drag", PPlanePlugin);
    PPlanePlugin.prototype = Object.create(mpld3.Plugin.prototype);
    PPlanePlugin.prototype.constructor = PPlanePlugin;
    PPlanePlugin.prototype.requiredProps = ["id", "idmpoint", 
                                            "idlinesla", "idlineslb", "idlinesra", "idlinesrb",  
                                            "idqlm", "idqmm", "idqrm", "idqone", "idqtwo", 
                                            "idrl0", "idrl1", "idrr0", "idrr1"];
    PPlanePlugin.prototype.defaultProps = {}
    function PPlanePlugin(fig, props){
        mpld3.Plugin.call(this, fig, props);
        mpld3.insert_css("#" + fig.figid + " path.dragging",
                         {"fill-opacity": "1.0 !important",
                          "stroke-opacity": "1.0 !important"});
    };

    // Call draw function, this function is being looped all the time
    PPlanePlugin.prototype.draw = function(){
        // Get elements into script variables
        var obj = mpld3.get_element(this.props.id);
        var midpoint = mpld3.get_element(this.props.idmpoint);
        var linesla = mpld3.get_element(this.props.idlinesla);
        var lineslb = mpld3.get_element(this.props.idlineslb);
        var linesra = mpld3.get_element(this.props.idlinesra);
        var linesrb = mpld3.get_element(this.props.idlinesrb);
        var qlm = mpld3.get_element(this.props.idqlm);
        var qmm = mpld3.get_element(this.props.idqmm);
        var qrm = mpld3.get_element(this.props.idqrm);
        var qone = mpld3.get_element(this.props.idqone);
        var qtwo = mpld3.get_element(this.props.idqtwo);
        var rl0 = this.props.idrl0;
        var rl1 = this.props.idrl1;
        var rr0 = this.props.idrr0;
        var rr1 = this.props.idrr1;
        
        // Set initial conditions for javascript calculations
        var qleft = obj.offsets[0];
        var qright = obj.offsets[1];
        var qmid = midpoint.offsets[0];
        var off = 13;
        // Calculate slopes for eigenlines
        var ml = rl1/rl0;
        var mr = rr1/rr0;
        
        // Main d3 drag function
        var drag = d3.behavior.drag()
            .origin(function(d) { return {x:obj.ax.x(d[0]),
                                          y:obj.ax.y(d[1])}; })                              
            .on("dragstart", dragstarted)
            .on("drag", dragged)
            .on("dragend", dragended);
    
        // Set elements of ql and qr points and call main drag function
        obj.elements()
           .data(obj.offsets)
           .style("cursor", "default")
           .call(drag);      

        // Begin drag function
        function dragstarted(d) {
          d3.event.sourceEvent.stopPropagation();
          d3.select(this).classed("dragging", true);
        }
    
        // The drag function called while dragging is happening (meat of code here)
        function dragged(d,i) {
          // Convert mouse coordinates in drag event (d3.event) to python coordinates d
          d[0] = obj.ax.x.invert(d3.event.x);
          d[1] = obj.ax.y.invert(d3.event.y);             
          // Move ql and qr stored in obj (they have been selected in drag)
          d3.select(this)
            .attr("transform", "translate(" + [d3.event.x, d3.event.y] + ")");
          // If obj corresponds to ql, move all the other left elements 
          if (i==0){
              // Move text marker
              qlm.elements().transition().duration(1)
                  .attr("transform", "translate(" + [d3.event.x + off, d3.event.y + off] + ")");
                  
              // Move eigenlines
              for (var ii=0; ii<2; ii++) {
                linesla.data[ii][1] = ml*(linesla.data[ii][0] - d[0]) + d[1];
                lineslb.data[ii][1] = mr*(lineslb.data[ii][0] - d[0]) + d[1]; }
                
              // In script calculations of middle state
              qleft = [d[0], d[1]];
              var det = rl0*rr1 - rr0*rl1;
              var alphal = (rr1*(qright[0] - qleft[0]) - rr0*(qright[1] - qleft[1]))/det
              qmid[0] = qleft[0] + alphal*rl0;
              qmid[1] = qleft[1] + alphal*rl1;
              var xx = obj.ax.x(qmid[0]);
              var yy = obj.ax.y(qmid[1]); 
             
                }
          // if element corresponds to qr
          else {
              // Move text marker
              qrm.elements().transition().duration(1)
                  .attr("transform", "translate(" + [d3.event.x + off, d3.event.y + off] + ")");
                  
              // Move eigenlines
              for (var ii=0; ii<2; ii++) {
                linesra.data[ii][1] = ml*(linesra.data[ii][0] - d[0]) + d[1];
                linesrb.data[ii][1] = mr*(linesrb.data[ii][0] - d[0]) + d[1]; }
                
              // In script calculations of middle state    
              qright = [d[0], d[1]];
              var det = rl0*rr1 - rr0*rl1;
              var alphal = (rr1*(qright[0] - qleft[0]) - rr0*(qright[1] - qleft[1]))/det
              qmid[0] = qleft[0] + alphal*rl0;
              qmid[1] = qleft[1] + alphal*rl1;
              var xx = obj.ax.x(qmid[0]);
              var yy = obj.ax.y(qmid[1]);
                }
              // Update middle state point and marker position
              midpoint.elements().transition().duration(5)
                .attr("transform", "translate(" + [xx, yy] + ")");
              qmm.elements().transition().duration(5)
                .attr("transform", "translate(" + [xx + 0.7*off, yy + 0.7*off] + ")"); 
                
              // Update eigenlines
              linesla.elements().transition().duration(5)
                  .attr("d", linesla.datafunc(linesla.data));
              lineslb.elements().transition().duration(5)
                  .attr("d", lineslb.datafunc(lineslb.data));
              linesra.elements().transition().duration(5)
                  .attr("d", linesra.datafunc(linesra.data));
              linesrb.elements().transition().duration(5)
                  .attr("d", linesrb.datafunc(linesrb.data));
                
              // Update subplots of q1 and q2
              qone.data[0][1] = qleft[0];
              qone.data[1][1] = qleft[0];
              qone.data[2][1] = qmid[0];
              qone.data[3][1] = qmid[0];
              qone.data[4][1] = qright[0];
              qone.data[5][1] = qright[0];
              qtwo.data[0][2] = qleft[1];
              qtwo.data[1][2] = qleft[1];
              qtwo.data[2][2] = qmid[1];
              qtwo.data[3][2] = qmid[1];
              qtwo.data[4][2] = qright[1];
              qtwo.data[5][2] = qright[1];
              qone.elements().transition().duration(5)
                .attr("d", qone.datafunc(qone.data));
              qtwo.elements().transition().duration(5)
                .attr("d", qtwo.datafunc(qtwo.data));  
        }
        // End dragging
        function dragended(d) {
          d3.select(this).classed("dragging", false);
        }
    }
    mpld3.register_plugin("drag", PPlanePlugin); 
    """

    def __init__(self, points, midpoint, 
                 linesla, lineslb, linesra, linesrb, 
                 qlmarker, qmmarker, qrmarker, 
                 qone, qtwo, rl0, rl1, rr0, rr1):
        if isinstance(points, mpl.lines.Line2D):
            suffix = "pts"
        else:
            suffix = None

        self.dict_ = {"type": "drag",
                      "id": utils.get_id(points, suffix),
                      "idmpoint": utils.get_id(midpoint,suffix),
                      "idlinesla": utils.get_id(linesla),
                      "idlineslb": utils.get_id(lineslb),
                      "idlinesra": utils.get_id(linesra),
                      "idlinesrb": utils.get_id(linesrb),
                      "idqlm": utils.get_id(qlmarker,suffix),
                      "idqmm": utils.get_id(qmmarker,suffix),
                      "idqrm": utils.get_id(qrmarker,suffix),
                      "idqone": utils.get_id(qone),
                      "idqtwo": utils.get_id(qtwo),
                      "idrl0": rl0,
                      "idrl1": rl1,
                      "idrr0": rr0,
                      "idrr1": rr1}


def linear_phase_plane(ql=np.array([-2.0, 2.0]),qr=np.array([0.0, 3.0]), 
                       r1=None, r2=None, lam1=None, lam2=None, 
                       q1min=-5, q1max=5, q2min =-5, q2max=5, domain=5, time=1.0, 
                       title1=None, title2=None):
    # Create figure
    # Create a figure
    fig, ax = plt.subplots(1,3, figsize=(13.5, 5.0))
    fig.subplots_adjust(left=0.05, right=0.9, bottom=0.2, top=0.9,
                        hspace=0.3, wspace=0.15)

    #Create subfigure ax1 for phaseplane
    #ax[0] = fig.add_subplot(2,2,1)

    # Calculate dq from ql and qr
    dq = np.array([qr[0]-ql[0], qr[1]-ql[1]])

    # If eigenvectors not specified, default to Acoustics
    rho0 = 1.0
    zz = 2.0
    if r1 is None:
        r1 = np.array([-zz, 1])
    if r2 is None:
        r2 = np.array([zz,  1])
    if lam1 is None:
        lam1 = -zz/rho0
    if lam2 is None:
        lam2 = zz/rho0
    
    # Choose left and right eigenvectors
    if (lam1 < lam2):
        laml = lam1; lamr = lam2
        rl = r1; rr = r2 
    else:
        laml = lam2; lamr = lam1
        rl = r2; rr = r1      
        
    # Assign titles to solution plots
    if title1 is None:
        title1 = "q1"
    if title2 is None:
        title2 = "q2"
    
    #Plot eigenlines across ql and qr
    eps = 0.00000000001 # To avoid bugs in D3
    ml = rl[1]/rl[0]
    mr = rr[1]/rr[0] + eps
    linesla = ax[0].plot([q1min,q1max],[ml*(q1min - ql[0]) + ql[1],ml*(q1max - ql[0]) + ql[1]], '-k')
    lineslb = ax[0].plot([q1min-eps,q1max+eps],[mr*(q1min - ql[0]) + ql[1],mr*(q1max - ql[0]) + ql[1]], '-k')
    linesra = ax[0].plot([q1min-2*eps,q1max+2*eps],[mr*(q1min - qr[0]) + qr[1],mr*(q1max - qr[0]) + qr[1]], '-k')
    linesrb = ax[0].plot([q1min-3*eps,q1max+3*eps],[ml*(q1min - qr[0]) + qr[1],ml*(q1max - qr[0]) + qr[1]], '-k')

    # Plot ql and qr
    points = ax[0].plot([ql[0],qr[0]], [ql[1], qr[1]], 'ok', alpha=0.7, markersize=10, markeredgewidth=1)
    data = ["q_l", "q_r"]
    offset = 0.4*0.5*(q1max-q1min)/5.0
    qlmarker = ax[0].plot(ql[0] + offset, ql[1] - offset, 'ok', marker=(r"$ q_l $"), markersize=15)
    qrmarker = ax[0].plot(qr[0] + offset, qr[1] - offset, 'ok', marker=(r"$ q_r $"), markersize=15)

    #Plot midpoint
    det = rl[0]*rr[1] - rr[0]*rl[1]
    alL = (rr[1]*dq[0] - rr[0]*dq[1])/det
    qm = ql + alL*rl 
    midpoint = ax[0].plot(qm[0],qm[1],'ok', alpha=0.9, markersize=8, markeredgewidth=1)
    qmmarker = ax[0].plot(qm[0]+offset,qm[1]-0.7*offset, 'k',marker=(r"$ q_m $"),markersize=20)

    # Set axis 1 properties
    ax[0].set_title("Phase Plane", fontsize=18)
    ax[0].axis([q1min,q1max,q2min,q2max])
    ax[0].grid(alpha=0.1,color='k', linestyle='--')
    ax[0].set_xlabel(title1)
    ax[0].set_ylabel(title2)

    # Remove defult mpld3 plugins
    plugins.clear(fig)

    # Create solutionl line with six points where solution should be
    ctleft = laml*time
    ctright = lamr*time
    domain = max(domain,abs(ctleft)+1,abs(ctright)+1) # Readjust xdomain if necessarry
    xsol = np.array([-domain,ctleft,ctleft,ctright,ctright,domain])
    qsol1 = 1*xsol
    qsol1[0:2] = ql[0]
    qsol1[2:4] = qm[0]
    qsol1[4:6] = qr[0]

    # Set axis 2 properties
    ax[1].set_title(title1, fontsize=18)
    ax[1].axis([-domain,domain,q1min,q1max])
    ax[1].grid(alpha=0.1,color='k', linestyle='--')
    ax[1].set_xlabel('x')
    #ax[1].set_ylabel(title1)
    
    # Plot solution
    qone = ax[1].plot(xsol, qsol1, '-k', linewidth = 4, alpha = 1.0)

    #Create subfigure ax2 for solution plane
    #ax[2] = fig.add_subplot(2,2,2)
    # Create solutionl line with six points
    xsol2 = np.array([-domain,ctleft,ctleft,ctright,ctright,domain])
    qsol2 = 1*xsol2
    qsol2[0:2] = ql[1]
    qsol2[2:4] = qm[1]
    qsol2[4:6] = qr[1]

    # Set axis 2 properties
    ax[2].set_title(title2, fontsize=18)
    ax[2].axis([-domain,domain,q2min,q2max])
    ax[2].grid(alpha=0.1,color='k', linestyle='--')
    ax[2].set_xlabel('x')
    #ax[2].set_ylabel(title2)

    # Plot solution
    qtwo = ax[2].plot(xsol2, qsol2, '-k', linewidth = 4, alpha = 1.0)

    # Call mpld3 custom PPLane plugin to interact with plot
    plugins.connect(fig, PPlanePlugin(points[0],midpoint[0],
                                      linesla[0],lineslb[0],linesra[0],linesrb[0],
                                      qlmarker[0],qmmarker[0],qrmarker[0],qone[0],qtwo[0],
                                      rl[0],rl[1],rr[0],rr[1]))
    
    return fig


### Tests for interactive apps
def test_interactive_shallowPP(save_to_html=False):
    # Define left and right state (h,hu)
    ql = np.array([3.0, 5.0])
    qr = np.array([3.0, -5.0])
    # Define optional parameters (otherwise chooses default values)
    plotopts = {'g':1.0, 'time':2.0, 'tmax':5, 'hmax':10, 'humin':-15, 'humax':15}
    # Call interactive function (can be called without any argument)
    pt = shallow_water(ql,qr,**plotopts)
    if save_to_html:
        mpld3.save_html(pt, "test_shallow.html")
    mpld3.show()

def test_interactive_linearPP(save_to_html=False):
    ## Define left and right state 
    ql = np.array([-2.0, 2.0]) 
    qr = np.array([0.0, -3.0])
    # Define two eigenvectors and eigenvalues (acoustics)
    zz = 2.0
    rho0 = 1.0
    r1 = np.array([zz,1.0])
    r2 = np.array([-zz,1.0])
    lam1 = zz/rho0
    lam2 = -zz/rho0
    plotopts={'q1min':-5, 'q1max':5, 'q2min':-5, 'q2max':5, 'domain':5, 'time':1, 
            'title1':"Pressure", 'title2':"Velocity"}
    pt = linear_phase_plane(ql,qr,r1,r2,lam1,lam2,**plotopts)
    if save_to_html:
        mpld3.save_html(pt, "test_linearPP.html")
    mpld3.show()



"""
Tools for Lagrangian gauges (particle tracking).
First attempt, could be improved.
"""


from __future__ import absolute_import
from __future__ import print_function

import numpy
from clawpack.visclaw import gaugetools
from clawpack.pyclaw import gauges
from clawpack.amrclaw.data import GaugeData

def read_gauges(gaugenos='all', outdir=None):
    
    if (type(gaugenos) is int) or (type(gaugenos) is str):
        gaugenos = [gaugenos]  # single gaugeno passed in
        
    gd = GaugeData(outdir)
    gd.read()
    
    gauge_solutions = {}
    
    for gaugeno in gd.gauge_numbers:
        if (gaugeno in gaugenos) or (gaugenos[0] == 'all'):
            g = gauges.GaugeSolution(gaugeno, outdir)
            gauge_solutions[gaugeno] = g

    if 0:
        # for debugging:
        gaugenos_lagrangian = [k for k in gauge_solutions.keys() \
                    if gauge_solutions[k].gtype=='lagrangian']
        gaugenos_stationary = [k for k in gauge_solutions.keys() \
                    if gauge_solutions[k].gtype=='stationary']
        print('+++ Loaded %i stationary and %i Lagrangian gauges' \
            % (len(gaugenos_stationary),len(gaugenos_lagrangian)))
            
    return gauge_solutions
        
def check_gaugenos_input(gauge_solutions, gaugenos):
    """
    If `gaugenos` is 'all', replace by list of lagrangian gauges
    If `gaugenos` is a single integer, turn into list
    Then check if each gaugeno is lagrangian, and add to output
        list only if lagrangian
    Return modified list of lagrangian gaugenos
    """
    gaugenos_lagrangian = []
    gaugenos_stationary = []
    
    if gaugenos=='all':
        gaugenos = gauge_solutions.keys()    
    elif type(gaugenos) is int:
        gaugenos = [gaugenos]  # single gaugeno passed in
        
    for gaugeno in gaugenos:
        if gauge_solutions[gaugeno].gtype == 'lagrangian':
            gaugenos_lagrangian.append(gaugeno)
        else:
            gaugenos_stationary.append(gaugeno)
            
    if len(gaugenos_lagrangian) == 0:
        print('*** Warning, no lagrangian gauges in gaugenos')
        
    return gaugenos_lagrangian
               
                           
def interp_particles(gauge_solutions, t, gaugenos='all', extend='neither'):
    """
    Interpolate (x,y) to the given time t for each specified gauge.
    Returns a dictionary of results indexed by gauge number.
    """
        
    gaugenos = check_gaugenos_input(gauge_solutions, gaugenos)
    
    particle_positions = {}
    for gaugeno in gaugenos:
        g = gauge_solutions[gaugeno]
        if g.particle_path is None: 
            particle_positions[gaugeno] = (numpy.nan, numpy.nan)
            continue # no gauge data yet, go on to next gauge
        tg = g.particle_path[:,0]
        xg = g.particle_path[:,1]
        yg = g.particle_path[:,2]
        if t <= tg[0]:
            if extend in ['neither', 'max']:
                x = numpy.nan
                y = numpy.nan
            elif extend in ['min', 'both']:
                x = xg[0]
                y = yg[0]
            else:
                raise ValueError('Unrecognized extend')
        elif t > tg[-1]:
            if extend in ['neither', 'min']:
                x = numpy.nan
                y = numpy.nan
            elif extend in ['max', 'both']:
                x = xg[-1]
                y = yg[-1]
            else:
                raise ValueError('Unrecognized extend')
        else:
            i = numpy.where(tg<=t)[0][-1]
            if i > len(tg)-2: 
                i = len(tg)-2
            if i<len(tg)-1:
                alf = (t-tg[i])/(tg[i+1]-tg[i])
                x = xg[i] + alf*(xg[i+1]-xg[i])
                y = yg[i] + alf*(yg[i+1]-yg[i])
            else:
                x = xg[-1]
                y = yg[-1]
        particle_positions[gaugeno] = (x,y)
    return particle_positions


def plot_particles(gauge_solutions, t, gaugenos='all', kwargs_plot=None,
                   extend='neither'):
    """
    Plot particle locations as points for some set of gauges.
    """
    from matplotlib import pyplot as plt

    gaugenos = check_gaugenos_input(gauge_solutions, gaugenos)
        
    if kwargs_plot is None:
        kwargs_plot = {'marker':'o','markersize':2,'color':'k'}
        
    pp = interp_particles(gauge_solutions, t, gaugenos, extend)
    for k in pp.keys():
        x,y = pp[k]
        plt.plot([x],[y],**kwargs_plot)
        

def plot_paths(gauge_solutions, t1=None, t2=None, gaugenos='all', 
               kwargs_plot=None, extend='neither'):
    """
    Plot the particle path for a set of gauges over some time interval.
    """
    from matplotlib import pyplot as plt
    
    
    if kwargs_plot is None:
        kwargs_plot = {'linestyle':'-','linewidth':0.7,'color':'k'}

    gaugenos = check_gaugenos_input(gauge_solutions, gaugenos)
    
    for gaugeno in gaugenos:
        g = gauge_solutions[gaugeno]
        if g.particle_path is None: 
            continue # no gauge data yet, go on to next gauge
        tg = g.particle_path[:,0]
        xg = g.particle_path[:,1]
        yg = g.particle_path[:,2]
        if 1:
            if t1 is None:
                i1 = 0
                t1 = tg[0]
            else:
                try:
                    i1 = numpy.where(tg>=t1)[0][0]
                except:
                    i1 = None
            if t2 is None:
                i2 = -1
                t2 = tg[-1]
            else:
                try:
                    i2 = numpy.where(tg<=t2)[0][-1]
                except:
                    i2 = None
            
            if (i1 is not None) and (i2 is not None):
                pp1 = interp_particles(gauge_solutions, t1, gaugeno, extend)
                pp2 = interp_particles(gauge_solutions, t2, gaugeno, extend)
                xp = numpy.hstack((pp1[gaugeno][0], xg[i1:i2], pp2[gaugeno][0]))
                yp = numpy.hstack((pp1[gaugeno][1], yg[i1:i2], pp2[gaugeno][1]))
                plt.plot(xp, yp, **kwargs_plot)
            else:
                pass # no points in this time range




#!/usr/bin/env python
# encoding: utf-8
r"""
Compressible Euler flow over a forward-facing step
==================================================

Solve the Euler equations of compressible fluid dynamics in 2D:

.. math::
    \rho_t + (\rho u)_x + (\rho v)_y = 0 \\
    (\rho u)_t + (\rho u^2 + p)_x + (\rho uv)_y = 0 \\
    (\rho v)_t + (\rho uv)_x + (\rho v^2 + p)_y = 0 \\
    E_t + (u (E + p) )_x + (v (E + p))_y = 0.


Here :math:`\rho` is the density, (u,v) is the velocity, and E is the total energy.

The problem involves a shock wave impacting a forward-facing step.  A
stationary shock forms.  When using a Roe solver, there is a carbuncle
instability.

This example demonstrates how to include internal boundaries within the
domain.  An aux array field is used to indicate which cells are inside (1)
or outside (0) of the domain.  A special Riemann solver enforces reflecting
boundary conditions at the internal boundaries.  This could be modified
to handle other internal boundary geometries, as long as they are
aligned with the grid.
"""

gamma = 1.4 # Ratio of specific heats

def incoming_shock(state,dim,t,qbc,auxbc,num_ghost):
    """
    Incoming shock at left boundary.
    """
    rho = 1.4
    u   = 3.0
    v   = 0.0
    p   = 1.0
    T   = 2*p/rho

    for i in range(num_ghost):
        qbc[0,i,...] = rho
        qbc[1,i,...] = rho*u
        qbc[2,i,...] = rho*v
        qbc[3,i,...] = rho*(5.0/4 * T + (u**2+v**2)/2.)


def setup(use_petsc=False,solver_type='classic', outdir='_output', kernel_language='Fortran',
          disable_output=False, mx=320, my=80, tfinal=2.0, num_output_times = 10):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    from clawpack import riemann

    if solver_type=='sharpclaw':
        #solver = pyclaw.SharpClawSolver2D(euler_with_boundaries)
        raise Exception('This problem does not currently work with SharpClaw')
    else:
        solver = pyclaw.ClawSolver2D(riemann.euler_hlle_with_walls_2D)
        solver.dimensional_split = True

    solver.num_eqn = 4
    solver.num_waves = 4
    num_aux = 1

    x = pyclaw.Dimension(0.0,3.0,mx,name='x')
    y = pyclaw.Dimension(0.0,1.0,my,name='y')
    domain = pyclaw.Domain([x,y])

    state = pyclaw.State(domain,solver.num_eqn,num_aux)
    state.problem_data['gamma']= gamma

    solver.user_bc_lower = incoming_shock

    solver.bc_lower[0]=pyclaw.BC.custom
    solver.bc_upper[0]=pyclaw.BC.extrap
    solver.bc_lower[1]=pyclaw.BC.wall
    solver.bc_upper[1]=pyclaw.BC.wall

    solver.aux_bc_lower[0]=pyclaw.BC.extrap
    solver.aux_bc_upper[0]=pyclaw.BC.extrap
    solver.aux_bc_lower[1]=pyclaw.BC.wall
    solver.aux_bc_upper[1]=pyclaw.BC.wall

    rho = 1.4
    u = 3.0
    v = 0.
    p = 1.0
    T = 2*p/rho

    state.q[0,...] = rho
    state.q[1,...] = rho*u
    state.q[2,...] = rho*v
    state.q[3,...] = rho*(5.0/4 * T + (u**2+v**2)/2.)

    x,y = domain.grid.p_centers

    # Set aux values to zero inside step, to unity elsewhere
    state.aux[0,...]= 1 - (x > 0.6)*(y < 0.2)

    claw = pyclaw.Controller()
    claw.solution = pyclaw.Solution(state,domain)
    claw.solver = solver

    claw.keep_copy = True
    if disable_output:
        claw.output_format = None
    claw.tfinal = tfinal
    claw.num_output_times = num_output_times
    claw.outdir = outdir
    claw.setplot = setplot

    return claw

    
def setplot(plotdata):
    "Plot solution using VisClaw." 
    from clawpack.visclaw import colormaps

    plotdata.clearfigures()  # clear any old figures,axes,items data
    
    # Density plot
    plotfigure = plotdata.new_plotfigure(name='Density', figno=0)

    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Density'
    plotaxes.scaled = True      # so aspect ratio is 1
    plotaxes.afteraxes = fill_step
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmin = 1
    plotitem.pcolor_cmax = 7.0
    plotitem.add_colorbar = True
    

    # Schlieren plot of density
    plotfigure = plotdata.new_plotfigure(name='Schlieren', figno=1)

    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Density (Schlieren)'
    plotaxes.scaled = True      # so aspect ratio is 1
    
    plotaxes.afteraxes = fill_step
    plotitem = plotaxes.new_plotitem(plot_type='2d_schlieren')
    plotitem.schlieren_cmin = 0.
    plotitem.schlieren_cmax = 3.
    plotitem.plot_var = 0
    plotitem.add_colorbar = False
    
    return plotdata
   
def fill_step(_):
    "Fill the step area with a black rectangle."
    import matplotlib.pyplot as plt
    rectangle = plt.Rectangle((0.6,0.0),2.4,0.2,color="k",fill=True)
    plt.gca().add_patch(rectangle)


if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup,setplot)




from __future__ import print_function
from pylab import *
from scipy.interpolate import interp1d
import os

savefig_ext = '.png'
figdir = './figures'
os.system('mkdir -p %s' % figdir)

def save_figure(fname):
    """Save figure to figdir with desired extension"""
    full_fname = os.path.join(figdir,fname) + savefig_ext
    savefig(full_fname, bbox_inches='tight')
    print('Created %s' % full_fname)

outdir = '_output'
final_frame = 75
dp = loadtxt(outdir + '/fort.q%s' % str(final_frame).zfill(4), skiprows=5)
#dn = loadtxt('eta_neg.txt', skiprows=5)
g = loadtxt(outdir + '/celledges.data',skiprows=1)
xg = (g[:-1,0]+g[1:,0])/2.

eta_pos = dp[:,2]
#eta_neg = dn[:,2]
eta0 = interp1d(xg,eta_pos,fill_value="extrapolate")

def make_plot(frameno):
    figure(5, figsize=(7,6))
    clf()
    dn = loadtxt(outdir + '/fort.q%s' % str(frameno).zfill(4), skiprows=5)
    eta_neg = -dn[:,2]
    eta1 = interp1d(xg,eta_neg,fill_value="extrapolate")
    xx = linspace(-300e3,50e3,4000)
    subplot(211)
    plot(xx/1000,eta0(xx),'b')
    plot(xx/1000,eta1(xx),'g')
    title('Solution with positive step (blue) and negative step (green)')
    grid(True)
    xticks()
    ylabel('meters')
    subplot(212)
    plot(xx/1000, eta0(xx)+eta1(xx), 'r')
    title('Solution with square pulse (sum of step solutions)')
    grid(True)
    xlabel('kilometers')
    ylabel('meters')
    tight_layout()
    fname = 'steps_%s' % str(frameno).zfill(4)
    save_figure(fname)

if __name__=='__main__':
    make_plot(60)
    make_plot(70)



#!/usr/bin/env python
# encoding: utf-8

""" 
Test problem demonstrating a 1D shocktube in a 3D domain. 

This script evolves the 3D Euler equations.
The primary variables are: 
    density (rho), x,y, and z momentum (rho*u,rho*v,rho*w), and energy.
"""
from clawpack import riemann
from clawpack.riemann.euler_3D_constants import density, x_momentum, \
                y_momentum, z_momentum, energy, num_eqn

gamma = 1.4 # Ratio of Specific Heats

def shocktube(kernel_language='Fortran', solver_type='classic',
              use_petsc=False, outdir='shocktube_output', output_format='hdf5',
              disable_output=False, mx=10, my=10, mz=128, tfinal=1.0,
              num_output_times=10):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw
        
    if solver_type == 'classic':
        solver = pyclaw.ClawSolver3D(riemann.euler_3D)
        solver.dimensional_split = True
        solver.limiters = pyclaw.limiters.tvd.MC
        solver.cfl_max = 1.0
        solver.cfl_desired = 0.80
    elif solver_type == 'sharpclaw':
        solver = pyclaw.SharpClawSolver3D(riemann.euler_3D)
    else:
        raise Exception('Unrecognized solver_type.')

    domain = pyclaw.Domain((-1.,-1.,-1.),(1.,1.,1.),(mx,my,mz))

    state = pyclaw.State(domain,num_eqn)
    state.problem_data['gamma']  = gamma
    
    grid = state.grid
    
    X,Y,Z = grid.p_centers

    pressure = 3.*(Z<=0) + 1.*(Z>0)
    state.q[density   ,:,:,:] = 3.*(Z<=0) + 1.*(Z>0)
    state.q[x_momentum,:,:,:] = 0.
    state.q[y_momentum,:,:,:] = 0.
    state.q[z_momentum,:,:,:] = 0.
    state.q[energy    ,:,:,:] = pressure/(gamma - 1.)

    solver.all_bcs = pyclaw.BC.extrap

    claw = pyclaw.Controller()
    claw.solution = pyclaw.Solution(state,domain)
    claw.solver = solver
    claw.output_format = output_format
    claw.keep_copy = True
    if disable_output:
        claw.output_format = None
    claw.tfinal = tfinal
    claw.num_output_times = num_output_times
    claw.outdir = outdir

    return claw

# __main__()
if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(shocktube)



r"""
fgmax_tools module: $CLAW/geoclaw/src/python/geoclaw/fgmax_tools.py

Tools to specify an fgmax grid for keeping track of maximum flow depth, etc.
and to read in the fgmax output after doing a GeoClaw run.

"""

import os
from numpy import sqrt, ma
import numpy

class FGmaxGrid(object):

    """
    New class introduced in 5.2.1 to keep store information both about the
    fgmax input data and the output generated by a GeoClaw run.
    """

    def __init__(self):

        super(FGmaxGrid, self).__init__()

        # GeoClaw input values:
        self.id = ''  # identifier, optional
        self.point_style = None
        self.npts = None
        self.nx = None
        self.ny = None
        self.n12 = None
        self.n23 = None
        self.tstart_max =  0.
        self.tend_max = 1.e10   # when to stop monitoring max values
        self.dt_check = 10.     # target time (sec) increment between updating
                                # max values
        self.min_level_check = None    # which levels to monitor max on
        self.interp_method = 0    # 0 for pw const, 1 for bilinear
        self.arrival_tol = 1.e-2       # tolerance for flagging arrival
        #self.input_file_name = 'fgmax.txt'  # file for GeoClaw input data
        self.fgno = None  # FG number
        self.xy_fname = None   # optional file name for separate list of points
                             # when point_style==0, distinct from header file
        self.write_xy_fname = False # controls whether xy_fname is created
                                    # by self.write_input_data, or only header

        # Other possible GeoClaw inputs:
        self.x = None
        self.y = None
        self.X = None
        self.Y = None
        self.Z = None  # for topo DEM values if available
        self.fgmax_point = None  # for point_style==4
        self.force_dry_init = None  # =1 if wet, =0 if dry
        self.dx = None
        self.dy = None

        # possible output values that may be available after run:
        self.outdir = '_output'    # where to find GeoClaw output fgmax*.txt
        self.level = None
        self.X = None
        self.Y = None
        self.Z = None
        self.dx = None
        self.dy = None
        self.B = None
        self.h = None
        self.h_time = None
        self.s = None
        self.s_time = None
        self.hs = None
        self.hs_time = None
        self.hss = None
        self.hss_time = None
        self.hmin = None
        self.hmin_time = None
        self.arrival_time = None

        # possible derived quantities of interest:
        self.dz = None
        self.B0 = None
        self.eta = None
        self.h_onshore = None
        self.label = ''  # text for legend



    def read_fgmax_grids_data(self, fgno, data_file='fgmax_grids.data'):
        """
        Read input info for fgmax grid number fgno from the data file
        fgmax_grids.data, which should have been created by setrun.py.
        This file now contains info about all fgmax grids.
        """
        with open(data_file) as filep:
            lines = filep.readlines()
        #print('+++ opened fgmax_new.data with %i lines' % len(lines))
        fgmax_input = None
        for lineno,line in enumerate(lines):
            if 'fgno' in line:
                if int(line.split()[0]) == fgno:
                    fgmax_input = lines[lineno+1:]
                    #print('Found line %i: %s' % (lineno,line))
                    break

        if fgmax_input is None:
            raise ValueError('fgmax grid fgno = %i not found in %s' \
                             % (fgno, data_file))

        self.fgno = fgno
        self.tstart_max = float(fgmax_input[0].split()[0])
        self.tend_max = float(fgmax_input[1].split()[0])
        self.dt_check = float(fgmax_input[2].split()[0])
        self.min_level_check = int(fgmax_input[3].split()[0])
        self.arrival_tol = float(fgmax_input[4].split()[0])
        self.interp_method = int(fgmax_input[5].split()[0])
        self.point_style = point_style = int(fgmax_input[6].split()[0])
        print('Reading input for fgno=%i, point_style = %i ' \
                % (self.fgno, self.point_style))
        if point_style == 0:
            self.npts = npts = int(fgmax_input[7].split()[0])
            if npts == 0:
                self.xy_fname = fgmax_input[8][1:-2]  # strip quotes
                xy = numpy.loadtxt(self.xy_fname, skiprows=1)
                self.X = xy[:,0]
                self.Y = xy[:,1]
                if xy.shape[1] > 2:
                    self.Z = xy[:,2]  # in case DEM values also stored in input file
                else:
                    self.Z = None
                self.npts = npts = len(self.X)
                print('Read %i x,y points from \n    %s' % (npts, self.xy_fname))
            else:
                X = []; Y = []
                for k in range(8,8+npts):
                    xk = float(fgmax_input[k].split()[0])
                    yk = float(fgmax_input[k].split()[1])
                    X.append(xk)
                    Y.append(yk)
                self.X = numpy.array(X)
                self.Y = numpy.array(Y)
        elif point_style == 1:
            self.npts = npts = int(fgmax_input[7].split()[0])
            self.x1 = float(fgmax_input[8].split()[0])
            self.y1 = float(fgmax_input[8].split()[1])
            self.x2 = float(fgmax_input[9].split()[0])
            self.y2 = float(fgmax_input[9].split()[1])
        elif point_style == 2:
            self.nx = nx = int(fgmax_input[7].split()[0])
            self.ny = ny = int(fgmax_input[7].split()[1])
            self.x1 = float(fgmax_input[8].split()[0])
            self.y1 = float(fgmax_input[8].split()[1])
            self.x2 = float(fgmax_input[9].split()[0])
            self.y2 = float(fgmax_input[9].split()[1])
        elif point_style == 3:
            self.n12 = n12 = int(fgmax_input[7].split()[0])
            self.n23 = n23 = int(fgmax_input[7].split()[1])
            self.x1 = float(fgmax_input[8].split()[0])
            self.y1 = float(fgmax_input[8].split()[1])
            self.x2 = float(fgmax_input[9].split()[0])
            self.y2 = float(fgmax_input[9].split()[1])
            self.x3 = float(fgmax_input[10].split()[0])
            self.y3 = float(fgmax_input[10].split()[1])
            self.x4 = float(fgmax_input[11].split()[0])
            self.y4 = float(fgmax_input[11].split()[1])
        elif point_style == 4:
            self.xy_fname = fgmax_input[7][1:-2]  # strip quotes
            ## Need to read in topotype 3 file and set self.npts
            # xy = numpy.loadtxt(self.xy_fname, skiprows=1)
            # self.X = xy[:,0]
            # self.Y = xy[:,1]
            # if xy.shape[1] > 2:
            #     self.Z = xy[:,2]  # in case DEM values also stored in input file
            # else:
            #     self.Z = None
            # self.npts = npts = len(self.X)
            # print('Read %i x,y points from \n    %s' % (npts, self.xy_fname))


    def write_to_fgmax_data(self, fid):
        """
        Write the fgmax grid data to the file specified by `fid`, normally
        the `fgmax_grids.data` file that is read in by the GeoClaw Fortran code.
        """

        print("\n---------------------------------------------- ")
        point_style = self.point_style
        if point_style not in [0,1,2,3,4]:
            raise NotImplementedError("make_fgmax not implemented for point_style %i" \
                % point_style)

        # write header, independent of point_style:
        #fid = open(self.input_file_name,'w')
        fid.write("\n")
        fid.write("%i                           # fgno\n" % self.fgno)
        fid.write("%16.10e            # tstart_max\n"  % self.tstart_max)
        fid.write("%16.10e            # tend_max\n"  % self.tend_max)
        fid.write("%16.10e            # dt_check\n" % self.dt_check)
        fid.write("%i %s              # min_level_check\n" \
                            % (self.min_level_check,12*" "))

        fid.write("%16.10e            # arrival_tol\n" % self.arrival_tol)
        fid.write("%i %s              # interp_method\n" \
                            % (self.interp_method,12*" "))
        fid.write("%i %s              # point_style\n" \
                            % (self.point_style,12*" "))

        print('fgmax grid %i has point_style = %i' % (self.fgno, point_style))

        if point_style == 0:
            if self.xy_fname is not None:
                fid.write("0         # npts==0 ==> points in this file:\n")
                fid.write("'%s'\n" % self.xy_fname)
                print("points should be in file:")
                print("   %s" % self.xy_fname)
                if self.write_xy_fname:
                    if self.Z is not None:
                        xydata = numpy.vstack([self.X,self.Y,self.Z]).T
                    else:
                        xydata = numpy.vstack([self.X,self.Y])
                    numpy.savetxt(self.xy_fname, xydata,
                                  header='%8i' % len(self.X),
                                  comments='', fmt='%24.14e')
            else:
                #print("+++ expecting xy_fname")
                # list of points
                npts = self.npts

                print("unstructured grid of %s points" % npts)

                fid.write("%i                 # npts\n" % (npts))
                for k in range(npts):
                    fid.write("%22.12f   %22.12f \n" % (self.X[k],self.Y[k]))
            #fid.close()


        elif point_style==1:
            # 1d transect of points
            x1,x2 = self.x1, self.x2
            y1,y2 = self.y1, self.y2
            # require self.npts to be set and don't set dx since
            # ambiguous for general transect in lat-long.
            if self.npts is None:
                raise ValueError('With point_style==1 must set set npts')
            else:
                npts = self.npts
                if self.dx is not None:
                    print("*** Warning: With point_style==1 cannot set dx")


            print("1d fixed grid with %s points" % npts)


            fid.write("%i                 # npts\n" % (npts))
            fid.write("%g   %g            # x1, y1\n" % (x1,y1))
            fid.write("%g   %g            # x2, y2\n" % (x2,y2))
            #fid.close()


            #print("Created file ", self.input_file_name)
            print("   specifying fixed grid with %i points equally spaced from " \
                    % npts)
            print("   (%g,%g)  to  (%g,%g)" % (x1,y1,x2,y2))


        if point_style == 2:
            # 2d grid of points
            x1,x2 = self.x1, self.x2
            y1,y2 = self.y1, self.y2
            if self.nx is None:
                dx = self.dx
                nx = int(round((x2-x1)/dx)) + 1
                if abs((nx-1)*dx + x1 - x2) > 1e-6:
                    print("Warning: abs((nx-1)*dx + x1 - x2) = ", \
                          abs((nx-1)*dx + x1 - x2))
                    print("         old x2: %22.16e" % x2)
                    x2 = x1 + dx*(nx-1)
                    print("         resetting x2 to %22.16e" % x2)
            else:
                nx = self.nx
                dx = (x2-x1)/(nx+1.)
                if self.dx is not None:
                    print("*** Warning: dx specified over-ridden by: ",dx)

            if self.ny is None:
                dy = self.dy
                if dy is None:
                    dy = dx
                ny = int(round((y2-y1)/dy)) + 1
                if abs((ny-1)*dy + y1 - y2) > 1e-6:
                    print("Warning: abs((ny-1)*dy + y1 - y2) = ", \
                          abs((ny-1)*dy + y1 - y2))
                    print("         old y2: %22.16e" % y2)
                    y2 = y1 + dy*(ny-1)
                    print("         resetting y2 to %22.16e" % y2)
            else:
                ny = self.ny
                dy = (y2-y1)/(ny+1.)
                if self.dy is not None:
                    print("*** Warning: dy specified over-ridden by: ",dy)


            npts = nx*ny


            fid.write("%i  %i %s          # nx,ny\n" \
                                % (nx,ny,10*" "))
            fid.write("%16.10e   %20.10e            # x1, y1\n" % (x1,y1))
            fid.write("%16.10e   %20.10e            # x2, y2\n" % (x2,y2))
            #fid.close()


            #print("Created file ", self.input_file_name)
            print("   specifying fixed grid with shape %i by %i, with  %i points" \
                    % (nx,ny,npts))
            print("   lower left  = (%15.10f,%15.10f)" % (x1,y1))
            print("   upper right = (%15.10f,%15.10f)" % (x2,y2))
            print("   dx = %15.10e,  dy = %15.10e" % (dx,dy))


        elif point_style==3:
            # arbitrary quadrilateral
            x1,x2 = self.x1, self.x2
            y1,y2 = self.y1, self.y2
            x3,x4 = self.x3, self.x4
            y3,y4 = self.y3, self.y4
            if self.n12 is None:
                raise NotImplementedError("Need to set n12 and n23")
            else:
                npts = self.n12 * self.n23

            fid.write("%i  %i %s          # self.n12,self.n23\n" \
                                % (self.n12,self.n23,10*" "))
            fid.write("%16.10e   %20.10e            # x1, y1\n" % (x1,y1))
            fid.write("%16.10e   %20.10e            # x2, y2\n" % (x2,y2))
            fid.write("%16.10e   %20.10e            # x3, y3\n" % (x3,y3))
            fid.write("%16.10e   %20.10e            # x4, y4\n" % (x4,y4))
            #fid.close()


            #print("Created file ", self.input_file_name)
            print("   specifying fixed grid as a quadrilateral")
            print("       %i by %i, with  %i points" \
                    % (self.n12,self.n23,npts))
            print("   corner 1 = (%15.10f,%15.10f)" % (x1,y1))
            print("   corner 2 = (%15.10f,%15.10f)" % (x2,y2))
            print("   corner 3 = (%15.10f,%15.10f)" % (x3,y3))
            print("   corner 4 = (%15.10f,%15.10f)" % (x4,y4))


        elif point_style == 4:
            if self.xy_fname is not None:
                fid.write("'%s'\n" % self.xy_fname)
                print("points should be in file:")
                print("   %s" % self.xy_fname)
            else:
                raise ValueError('for point_style==4, require xy_fname')

    def read_output(self, fgno=None, outdir=None, verbose=True, 
                    indexing='ij'):
        r"""
        Read the GeoClaw results on the fgmax grid numbered *fgno*.
        
        indexing='ij' gives backward compatibility.
           X[i,j],Y[i,j] corresponds to point x[i],y[j]

        Alternatively, can set indexing=='xy' so that X,Y and other
        arrays have same layout as topo arrays:
           X[j,i],Y[j,i] corresponds to point x[i],y[j]
        This is useful if you want to save the fgmax results in same format as 
        topofiles, using topotools.Topography.write().
           
        """

        if indexing == 'xy':
            reshape_order = 'C'
        elif indexing == 'ij':
            reshape_order = 'F'
        else:
            raise InputError("*** indexing must by 'xy' or 'ij'")
            
        if self.point_style is None:
            raise InputError("*** point_style is not set, need to read input?")
        point_style = self.point_style

        if fgno is not None:
            self.fgno = fgno
        if outdir is not None:
            self.outdir = outdir

        # Require new style in v5.7.0, e.g. fgmax0001.txt etc.
        fname = os.path.join(self.outdir, 'fgmax%s.txt' \
                % str(self.fgno).zfill(4))

        if not os.path.isfile(fname):
            raise IOError("File not found: %s" % fname)

        print("Reading %s ..." % fname)
        d = numpy.loadtxt(fname)

        if point_style == 4:
            self.npts = d.shape[0]
            print('point_style == 4, found %i points ' % self.npts)


        # new format in v5.7.0, includes column for B = topo from aux array
        cols_expected = [7,9,15]

        ncols = d.shape[1]

        if ncols not in cols_expected:
            raise IOError("*** Unexpected number of columns %s in file %s" \
                    % (ncols, fname))

        ind_s = None
        ind_hs = None
        ind_hss = None
        ind_hmin = None
        ind_h_time = None
        ind_s_time = None
        ind_hs_time = None
        ind_hss_time = None
        ind_hmin_time = None

        ind_x = 0
        ind_y = 1
        ind_level = 2
        ind_B = 3  # added in new fname style
        ind_h = 4
        if ncols == 7:
            ind_h_time = 5
            ind_arrival_time = 6
        elif ncols == 9:
            ind_s = 5
            ind_h_time = 6
            ind_s_time = 7
            ind_arrival_time = 8
        elif ncols == 15:
            ind_s = 5
            ind_hs = 6
            ind_hss = 7
            ind_hmin = 8
            ind_h_time = 9
            ind_s_time = 10
            ind_hs_time = 11
            ind_hss_time = 12
            ind_hmin_time = 13
            ind_arrival_time = 14


        if point_style in [0,1,4]:
            fg_shape = (self.npts,)
        elif point_style == 2:
            if indexing == 'xy':
                fg_shape = (self.ny,self.nx)
            else:
                fg_shape = (self.nx,self.ny)

        elif point_style == 3:
            if indexing == 'xy':
                fg_shape = (self.n23,self.n12)
            else:
                fg_shape = (self.n12,self.n23)
        else:
            raise NotImplementedError("Not implemented for point_style %s" \
                % point_style)

        X = numpy.reshape(d[:,0],fg_shape,order=reshape_order)
        Y = numpy.reshape(d[:,1],fg_shape,order=reshape_order)
        y0 = 0.5*(Y.min() + Y.max())   # mid-latitude for scaling plots
        h = numpy.reshape(d[:,ind_h],fg_shape,order=reshape_order)

        # AMR level used for each fgmax value:
        level = numpy.reshape(d[:,ind_level].astype('int'),fg_shape,
                              order=reshape_order)

        # Set B = topo array
        B = numpy.reshape(d[:,ind_B],fg_shape,order=reshape_order)

        mask = (h < -1e50)  # points that were never set
        B = ma.masked_where(mask, B)
        h = ma.masked_where(mask, h)

        def set_q_time(ind_q, ind_q_time):
            q = numpy.reshape(d[:,ind_q],fg_shape,order=reshape_order)
            q = ma.masked_where(mask,q)
            q_time = numpy.reshape(d[:,ind_q_time],fg_shape,order=reshape_order)
            q_time = ma.masked_where(mask, q_time)
            return q, q_time

        self.h, self.h_time = set_q_time(ind_h, ind_h_time)
        if ind_s:
            self.s, self.s_time = set_q_time(ind_s, ind_s_time)
        if ind_hs:
            self.hs, self.hs_time = set_q_time(ind_hs, ind_hs_time)
            self.hss, self.hss_time = set_q_time(ind_hss, ind_hss_time)
            self.hmin, self.hmin_time = set_q_time(ind_hmin, ind_hmin_time)

        # last column is arrival times:
        arrival_time = numpy.reshape(d[:,ind_arrival_time],
                                     fg_shape,order=reshape_order)
        arrival_time = ma.masked_where(arrival_time < -1e50, arrival_time)
        arrival_time = ma.masked_where(mask, arrival_time)
        self.arrival_time = arrival_time

        self.level = level
        self.X = X
        self.Y = Y
        self.B = B
        self.h = h

        # do not set these, leave for user to do as desired:
        if 0:
            self.B0 = B  ## SHOULD MODIFY BY dz!

            if self.force_dry_init is not None:
                self.h_onshore = ma.masked_where(self.force_dry_init==0, self.h)
            else:
                self.h_onshore = ma.masked_where(self.B0 < 0., self.h)

        if point_style==4:
            #print('Returning lists, convert to masked arrays based on input grid')
            #print('   using to_arrays() function')
            try:
                self.ps4_to_arrays(verbose=verbose)
            except:
                print('*** Problem converting from 1d lists to 2d arrays,\n' \
                      + '    Trying to map onto grid specified by:\n    ', \
                      self.xy_fname)
                raise

        if self.X.ndim==2:
            if indexing=='xy':
                self.x = self.X[0,:]
                self.y = self.Y[:,0]
            else:
                # for indexing=='ij' this fixes bug in v5.9.0 version:
                self.x = self.X[:,0]
                self.y = self.Y[0,:]
        else:
                self.x = self.X
                self.y = self.Y


    def bounding_box(self):
        """
        Return the bounding box of the grid as a list [x1,x2,y1,y2]
        """
        x1 = self.X.min()
        x2 = self.X.max()
        y1 = self.Y.min()
        y2 = self.Y.max()
        return [x1,x2,y1,y2]

    def ps4_to_arrays(self, verbose=True):
        """
        for point_style==4, convert lists of fgmax values into masked arrays
        based on the topo_style==3 file self.xy_fname that was used to specify
        the fgmax points in the GeoClaw run.
        """

        from numpy import ma
        assert self.point_style==4, '*** Requires point_style==4'

        if self.X.ndim==2 or self.Y.ndim==2:
            print('*** X and Y already 2d, not converting')
            return

        x_1d = self.X
        y_1d = self.Y

        if verbose:
            print('Will map fgmax points onto masked arrays defined by file:')
            print('     %s' % self.xy_fname)

        from clawpack.geoclaw import topotools
        pts_chosen = topotools.Topography(path=self.xy_fname, topo_type=3)
        X = pts_chosen.X
        Y = pts_chosen.Y
        mask = numpy.logical_not(pts_chosen.Z)
        x1 = X.min()
        y1 = Y.min()

        # possible arrays from GeoClaw output to convert:
        zarrays = ['level','B','h','h_time','s','s_time','hs','hs_time',\
                   'hss','hss_time','hmin','hmin_time','arrival_time']

        dx = X[0,1] - X[0,0]
        dy = Y[1,0] - Y[0,0]
        if verbose:
            print('Deduced dx = %g, dy = %g'  % (dx,dy))

        for attr in zarrays:
            z_1d = getattr(self, attr, None)
            if z_1d is None:
                if verbose: print('not converting attribute %s == None' % attr)
            else:
                Z = ma.masked_array(data=numpy.empty(X.shape), mask=True)
                for k in range(len(x_1d)):
                    i = int(round((x_1d[k]-x1)/dx))
                    j = int(round((y_1d[k]-y1)/dy))
                    Z[j,i] = z_1d[k]
                if 0:
                    if not numpy.alltrue(mask == Z.mask):
                        print('*** converting to arrays gave unexpected mask for')
                        print('    Z array =  %s' % attr)
                setattr(self, attr, Z)
                if verbose: print('converted %s to 2d array' % attr)

        self.X = X
        self.Y = Y



    def interp_dz(self,dtopo_path,dtopo_type):
        """
        Compute approximate values of deformation dz on X,Y grid using
        a specified dtopo file.
        Also calculates B0 = B - dz, attempting to recover the pre-event
        topography from the GeoClaw run topography stored in B.
        """
        from clawpack.geoclaw import dtopotools
        from scipy.interpolate import RegularGridInterpolator

        dtopo = dtopotools.DTopography(dtopo_path, dtopo_type=dtopo_type)
        x1d = dtopo.X[0,:]
        y1d = dtopo.Y[:,0]
        dtopo_func = RegularGridInterpolator((x1d,y1d), dtopo.dZ[-1,:,:].T,
                        method='linear', bounds_error=False, fill_value=0.)
        dz = dtopo_func(list(zip(numpy.ravel(self.X), numpy.ravel(self.Y))))
        self.dz = numpy.reshape(dz, self.X.shape)
        print('Over fgmax extent, min(dz) = %.2f m, max(dz) = %.2f m' \
             % (dz.min(), dz.max()))

        # leave this for user to do if desired:
        # self.B0 = self.B - self.dz


def adjust_fgmax_1d(x1_desired, x2_desired, x1_domain, dx):
    """
    Adjust the upper and lower limits of a grid so that equally spaced
    grid points with spacing `dx` lie exactly at cell centers, so that
    no interpolation is needed for fgmax values.  Note that parameter
    names refer to `x` limits, but works equally well for `y` values.

    :Input:
     - x1_desired, x2_desired: approximate desired limits of fgmax grid
     - x1_domain:  lower edge of computational domain
     - dx: Mesh spacing on fine grid that fgmax grid should conform to
    :Output:
     - x1_new, x2_new: limits to set so (x2-x1) is integer multiple
       of dx and points are at cell centers of computational grid
     - npoints: number of points to specify, so that
       `linspace(x1_new, x2_new, npoints) gives points with spacing `dx`.
    """

    i1 = numpy.floor((x1_desired-x1_domain - 0.5*dx)/dx)
    x1_new = x1_domain + (i1 + 0.5)*dx
    i2 = numpy.floor((x2_desired-x1_domain + 0.5*dx)/dx)
    x2_new = x1_domain + (i2 + 0.5)*dx
    npoints = int(i2 - i1) + 1
    return x1_new, x2_new, npoints


def adjust_fgmax_grid(x1_desired, x2_desired, x1_domain, dx,
                      y1_desired, y2_desired, y1_domain, dy=None, verbose=True):


    if dy == None:
        dy = dx

    x1_new, x2_new, nx = adjust_fgmax_1d(x1_desired, x2_desired, x1_domain, dx)
    y1_new, y2_new, ny = adjust_fgmax_1d(y1_desired, y2_desired, y1_domain, dy)

    if verbose:
        print("x:")
        print("  moved %17.12f to %17.12f by %g" % (x1_desired, x1_new, abs(x1_desired-x1_new)))
        print("  moved %17.12f to %17.12f by %g" % (x2_desired, x2_new, abs(x2_desired-x2_new)))
        print("y:")
        print("  moved %17.12f to %17.12f by %g" % (y1_desired, y1_new, abs(y1_desired-y1_new)))
        print("  moved %17.12f to %17.12f by %g" % (y2_desired, y2_new, abs(y2_desired-y2_new)))
        #print "  "
        #print "fg.nx = %g" % nx
        #print "fg.ny = %g" % ny
        #print "fg.x1 = %17.12f" % x1_new
        #print "fg.x2 = %17.12f" % x2_new
        #print "fg.y1 = %17.12f" % y1_new
        #print "fg.y2 = %17.12f" % y2_new
    return x1_new, x2_new, nx, y1_new, y2_new, ny




#!/usr/bin/env python
# encoding: utf-8

"""
2D shallow water equations.
"""
#===========================================================================
# Import libraries
#===========================================================================

import numpy as np
#from petclaw import plot
#import pdb  # Debugger

def init(state):
    # Initial solution
    # ================
    # Riemann states of the dam break problem
    radDam = 0.2
    hl = 2.
    ul = 0.
    vl = 0.
    hr = 1.
    ur = 0.
    vr = 0.
    
    x0=0.5
    y0=0.5
    xCenter = state.grid.x.centers
    yCenter = state.grid.y.centers
    
    Y,X = np.meshgrid(yCenter,xCenter)
    r = np.sqrt((X-x0)**2 + (Y-y0)**2)
    state.q[0,:,:] = hl*(r<=radDam) + hr*(r>radDam)
    state.q[1,:,:] = hl*ul*(r<=radDam) + hr*ur*(r>radDam)
    state.q[2,:,:] = hl*vl*(r<=radDam) + hr*vr*(r>radDam)


    
def shallow2D(use_petsc=False,outdir='./_output',solver_type='classic', disable_output=False):
    #===========================================================================
    # Import libraries
    #===========================================================================
    import numpy as np
    import clawpack.peanoclaw as peanoclaw
    import clawpack.riemann as riemann

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        import clawpack.pyclaw as pyclaw

    #===========================================================================
    # Setup solver and solver parameters
    #===========================================================================
    subdivisionFactor = 6
    if solver_type == 'classic':
        solver = pyclaw.ClawSolver2D(riemann.shallow_roe_with_efix_2D)
        solver.limiters = pyclaw.limiters.tvd.MC
        solver.dimensional_split=1
    elif solver_type == 'sharpclaw':
        solver = pyclaw.SharpClawSolver2D(riemann.shallow_roe_with_efix_2D)
    peanoSolver = peanoclaw.Solver(solver, (1./3.)/subdivisionFactor, init)
    
    solver.dt_initial = 1.0

    solver.bc_lower[0] = pyclaw.BC.wall
    solver.bc_upper[0] = pyclaw.BC.wall
    solver.bc_lower[1] = pyclaw.BC.wall
    solver.bc_upper[1] = pyclaw.BC.wall
    
    #===========================================================================
    # Initialize domain and state, then initialize the solution associated to the 
    # state and finally initialize aux array
    #===========================================================================

    # Domain:
    from clawpack.pyclaw import geometry
    print((geometry.__file__))
    xlower = 0.0
    xupper = 1.0
    mx = subdivisionFactor
    ylower = 0.0
    yupper = 1.0
    my = subdivisionFactor
    x = pyclaw.Dimension(xlower,xupper,mx,name='x')
    y = pyclaw.Dimension(ylower,yupper,my,name='y')
    domain = geometry.Domain([x,y])

    num_eqn = 3  # Number of equations
    state = pyclaw.State(domain,num_eqn)

    grav = 1.0 # Parameter (global auxiliary variable)
    state.problem_data['grav'] = grav
    
    #===========================================================================
    # Set up controller and controller parameters
    #===========================================================================
    claw = pyclaw.Controller()
    claw.tfinal = 0.1
    claw.solution = peanoclaw.solution.Solution(state,domain)
    claw.solver = peanoSolver
    claw.outdir = outdir
    if disable_output:
        claw.output_format = None
    claw.num_output_times = 5

    return claw


if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(shallow2D)
    print('Error: ', output)








#!/usr/bin/python
"""
okada
========================================================
This version of the Okada routines is deprecated but is
being kept around to test geoclaw/examples/tsunami/chile2010
Going forward, dtopotools.okadamap should be used instead.
========================================================

    python module for getting Okada maps.

    Okada model is a mapping from several fault parameters
    to a surface deformation.
    See Okada 1985, or Okada 1992, Bull. Seism. Soc. Am.

    some routines adapted from fortran routines written by
    Xiaoming Wang.

"""
import numpy
from numpy import *
from matplotlib import *
import matplotlib.mlab as mlab
import matplotlib.pyplot as pyplot
import string
from .datatools import *

#=================================================================================
def builddeffile (okadaparamfile,faultparamfile,outfile):

    faultparams=getokadaparams(okadaparamfile)
    faultparams.update(getfaultparams(faultparamfile))

    fid=open(outfile,'w')

    X=linspace(faultparams['xlower'],faultparams['xupper'],faultparams['mx'])
    Y=linspace(faultparams['ylower'],faultparams['yupper'],faultparams['my'])

    dZ=okadamap(faultparams,X,Y)
    ind=fixdata.findbadindices(dZ)
    if ind:
        dZ=fixdata.fillbaddata(dZ,ind)

    dZ = filtermask(dZ,faultparams)
    for jj in range(faultparams['my']):
        j=-1-jj
        for i in range(faultparams['mx']) :
            fid.write('%012.6e %012.6e %012.6e \n' % (X[i],Y[j],dZ[j,i]))

    fid.close()
    return

#=================================================================================
def builddynamicdeffile (okadaparamfile,faultparamfile,outfile,t0=0.0, tend=1.0, nt = 2):

    faultparams=getokadaparams(okadaparamfile)
    faultparams.update(getfaultparams(faultparamfile))

    fid=open(outfile,'w')

    X=linspace(faultparams['xlower'],faultparams['xupper'],faultparams['mx'])
    Y=linspace(faultparams['ylower'],faultparams['yupper'],faultparams['my'])

    T=linspace(t0,tend,nt)

    dZ=okadamap(faultparams,X,Y)
    ind=fixdata.findbadindices(dZ)
    if ind:
        dZ=fixdata.fillbaddata(dZ,ind)

    dZ = filtermask(dZ,faultparams)
    for it in T:
        alpha=(it-t0)/(tend-t0)
        for jj in range(faultparams['my']):
            j=-1-jj
            for i in range(faultparams['mx']) :
                fid.write('%012.6e %012.6e %012.6e %012.6e \n' % (it,X[i],Y[j],alpha*dZ[j,i]))

    fid.close()
    return

#=================================================================================
def getokadaparams (infile):

    """
    obtain parameters necessary for okada map from config file: infile

        file format:
        parameter names and values should appear on the same single line seperated by a space
    """

    keylist=["Focal_Depth","Fault_Length","Fault_Width","Dislocation","Strike_Direction", \
             "Dip_Angle","Slip_Angle","Epicenter_Latitude","Epicenter_Longitude"]

    okadaparams={}
    fid=open(infile,'r')
    keyleft=len(keylist)
    while keyleft> 0 :
        line=string.split(fid.readline())
        if line:
            if line[0] in keylist:
                okadaparams[line[0]]=float(line[1])
                keyleft=keyleft-1
            if line[1] in keylist:
                okadaparams[line[1]]=float(line[0])
                keyleft=keyleft-1

    for key in keylist :
        if not key in okadaparams:
            print(('ERROR: parameters for okada fault not fully specified in %s' % (infile)))
            exit

    fid.close()
    return okadaparams
    #end getokadaparams=============================================================

#===================================================================================
def getfaultparams (infile):

    """
    obtain params from a file that specify a fault grid from infile
    params are xlower,ylower,dx,dy,mx,my, OR
    xlower,ylower,xupper,yupper,mx,my

    file format:
        parameter names and values should appear on the same single line seperated by a space
    """

    keylist=["xlower","ylower","xupper","yupper","dx","dy","mx","my"]

    faultgridparams={}
    fid=open(infile,'r')
    keyleft=len(keylist)-2
    while keyleft> 0 :
        line=string.split(fid.readline())
        if line:
            if line[0] in keylist:
                faultgridparams[line[0]]=float(line[1])
                keyleft=keyleft-1
            if line[1] in keylist:
                faultgridparams[line[1]]=float(line[0])
                keyleft=keyleft-1

    faultgridparams['mx'] = int(faultgridparams['mx'])
    faultgridparams['my'] = int(faultgridparams['my'])

    if ('dx' in faultgridparams)& ('dy' in faultgridparams):
        faultgridparams['xupper'] = faultgridparams['xlower'] + faultgridparams['dx']*(faultgridparams['mx']-1)
        faultgridparams['yupper'] = faultgridparams['ylower'] + faultgridparams['dy']*(faultgridparams['my']-1)
    elif ('xupper' in faultgridparams)&('yupper' in faultgridparams):
        faultgridparams['dx'] = (faultgridparams['xupper']-faultgridparams['xlower'])/(faultgridparams['mx']-1)
        faultgridparams['dy'] = (faultgridparams['yupper']-faultgridparams['ylower'])/(faultgridparams['my']-1)
    else:
        print(('ERROR: parameters for fault grid not fully specified in %s' % (infile)))
        exit

    for key in keylist :
        if not key in faultgridparams:
            print(('ERROR: parameters for fault grid not fully specified in %s' % (infile)))
            exit

    fid.close()
    return faultgridparams
    #end getfaultparams===========================================================


#==================================================================================
def  okadamap(okadaparams,X,Y):

    """
    create displacement matrix dZ for a surface displacement
    over gridded region defined by X,Y, vectors of length nx,ny
    given okadaparams
    """

    xo=X[0]
    yo=Y[0]
    nx=len(X)
    ny=len(Y)

    zero = 0.0
    osixty = 0.016666666667
    rad = 0.01745329252
    rr = 6.378e6

    hh =  okadaparams["Focal_Depth"]
    l  =  okadaparams["Fault_Length"]
    w  =  okadaparams["Fault_Width"]
    d  =  okadaparams["Dislocation"]
    th =  okadaparams["Strike_Direction"]
    dl =  okadaparams["Dip_Angle"]
    rd =  okadaparams["Slip_Angle"]
#    yo =  okadaparams["Domain_Latitude"]
#    xo =  okadaparams["Domain_Longitude"]
    y0 =  okadaparams["Epicenter_Latitude"]
    x0 =  okadaparams["Epicenter_Longitude"]

    ang_l = rad*dl
    ang_r = rad*rd
    ang_t = rad*th
    halfl = 0.5*l

    #Calculate focal depth used for Okada's model
    hh = hh+w*sin(ang_l)
    #displacement due to different epicenter definition
    del_x = w*cos(ang_l)*cos(ang_t)
    del_y = w*cos(ang_l)*sin(ang_t)
    xl = rr*cos(rad*yo)*(x0-xo)*rad + del_x
    yl = rr*(y0-yo)*rad - del_y
    ds = d*cos(ang_r)
    dd = d*sin(ang_r)
    sn = sin(ang_l)
    cs = cos(ang_l)

    dZ=numpy.eye(ny,nx)

    # Vectorized by RJL, 12/10:
    x,y = meshgrid(X,Y)

    #!-----added by Xiaoming Wang, Nov 11 2006----
    xl = rr*cos(rad*y)*(x0-xo)*rad + del_x  # TAL - Fixed sign, 9/07
    #!---------------------------------------------
    yy = rr*(y-yo)*rad
    xx = rr*cos(rad*y)*(x-xo)*rad

    x1 = (xx-xl)*sin(ang_t)+(yy-yl)*cos(ang_t)
    x2 = (xx-xl)*cos(ang_t)-(yy-yl)*sin(ang_t)
    x2 = -x2
    x3 = zero
    p = x2*cs+hh*sn

    f1=strike_slip (x1,x2,x3,x1+halfl,p,ang_l,hh)
    f2=strike_slip (x1,x2,x3,x1+halfl,p-w,ang_l,hh)
    f3=strike_slip (x1,x2,x3,x1-halfl,p,ang_l,hh)
    f4=strike_slip (x1,x2,x3,x1-halfl,p-w,ang_l,hh)
    g1=dip_slip (x1,x2,x3,x1+halfl,p,ang_l,hh)
    g2=dip_slip (x1,x2,x3,x1+halfl,p-w,ang_l,hh)
    g3=dip_slip (x1,x2,x3,x1-halfl,p,ang_l,hh)
    g4=dip_slip (x1,x2,x3,x1-halfl,p-w,ang_l,hh)


    us = (f1-f2-f3+f4)*ds
    ud = (g1-g2-g3+g4)*dd

    dZ = (us+ud)

    return dZ
    #=========================================================================


#==============================================================================
def strike_slip (x1,x2,x3,y1,y2,dp,dd):
    """
    !.....Used for Okada's model
    !.. ..Methods from Yoshimitsu Okada (1985)
    !-----------------------------------------------------------------------
    """
    sn = sin(dp)
    cs = cos(dp)
    p = x2*cs + dd*sn
    q = x2*sn - dd*cs
    d_bar = y2*sn - q*cs
    r = sqrt(y1**2 + y2**2 + q**2)
    xx = sqrt(y1**2 + q**2)
    a4 = 0.5*1/cs*(log(r+d_bar) - sn*log(r+y2))
    f = -(d_bar*q/r/(r+y2) + q*sn/(r+y2) + a4*sn)/(2.0*3.14159)

    return f
    #==========================================================================


#==============================================================================
def dip_slip (x1,x2,x3,y1,y2,dp,dd):
    """
    !.....Based on Okada's paper (1985)
    !.....Added by Xiaoming Wang
    !-----------------------------------------------------------------------
    """
    sn = sin(dp)
    cs = cos(dp)

    p = x2*cs + dd*sn
    q = x2*sn - dd*cs
    d_bar = y2*sn - q*cs;
    r = sqrt(y1**2 + y2**2 + q**2)
    xx = sqrt(y1**2 + q**2)
    a5 = 0.5*2/cs*arctan((y2*(xx+q*cs)+xx*(r+xx)*sn)/y1/(r+xx)/cs)
    f = -(d_bar*q/r/(r+y1) + sn*arctan(y1*y2/q/r) - a5*sn*cs)/(2.0*3.14159)

    return f
    #===========================================================================


#================================================================================
def filtermask (dZ,faultparams):
    """
    borrowed from code written by Xiaoming Wang and Tom Logan at ARSC

    !.....Filter the deformation using a circular mask centered
    !.....at the epicenter using a calculated radius
    !.....Removes small numerical artifacts away from the epicenter
    """
    filterindices=[]

    osixty = 0.016666666667
    rad = 0.01745329252
    rr = 6.378e6

    xo = faultparams['xlower']
    yo = faultparams['ylower']
    nx = faultparams['mx']
    ny = faultparams['my']
    spacing = faultparams['dx']

    x0 = faultparams['Epicenter_Longitude']
    y0 = faultparams['Epicenter_Latitude']
    l =  faultparams['Fault_Length']
    w =  faultparams['Fault_Width']
    dl = faultparams['Dip_Angle']


    ang_l = rad*dl # convert degree to radian

    #!-- epicenter in pixels -----------
    ypix = (y0-yo)/spacing
    xpix = (x0-xo)/spacing

    #!-- conversion from meters to pixels ---
    tmpd=spacing*rad
    xdist = tmpd*rr

    #!-- size of the fault in pixels --------
    npix_x = l/xdist
    npix_y = w/xdist

    #!-- set the range (radius) of the filter circle --------
    #!----- for small dip angles, use the length and width --
    #!----- for larger dip angles, use only the length ------

    if dl<30.0:
        drange = 1.5 * cos(ang_l)*sqrt(npix_x*npix_x+npix_y*npix_y)
    else:
        drange = 1.2 * npix_x

    print(("Filtering deformation using a circle of radius %s" % (drange)))

    #!-- Create the filtering mask ----------
    for i in range(nx):
        for j in range(ny) :
            dist = sqrt((i+1-xpix)**2+(j+1-ypix)**2)
            if dist > drange :
                filterindices.append((j,i))

    #!-- apply the filter to the actual deformation ------
    dZ = fixdata.filterdata(dZ,filterindices,radius=2)

    return dZ



num_waves = 4
num_eqn   = 4
num_aux   = 5

# Conserved quantities
depth_1 = 0
momentum_1 = 1
depth_2 = 2
momentum_2 = 3

# Auxiliary variables
bathymetry = 0
wind = 1
h_hat_1 = 2
h_hat_2 = 3
kappa = 4



"""
Make piecewise linear topography for wave tank.
"""

from pylab import *
from clawpack.geoclaw import nonuniform_grid_tools

xlower = -160.8
xupper = 40.

xzpairs = [(-160.8, -4),       # left edge
           (-125.5, -4),       # start of first slope
           (-90,    -0.45),    # start of beach
           (  0,     0),       # shore
           ( 40,     2)]       # right edge

topo_fcn = nonuniform_grid_tools.make_pwlin_topo_fcn(xzpairs)

mx = 8000
hmin = 0.05  # use uniform grid in shallower water
#hmin = 5  # try forcing a uniform grid everywhere
#hmin = 1  # try uniform grid near start of shelf

nonuniform_grid_tools.make_celledges_cfl(xlower, xupper, mx, topo_fcn,
        hmin, fname='celledges.data', plot_topo=True)




"""Generate Fortran WENO kernels with PyWENO."""

# Copyright (c) 2011, 2014 Matthew Emmett.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import pyweno

kernel = pyweno.kernels.KernelGenerator('fortran')

#### set fortran templatese to match clawpack

template = '''
subroutine {function}(q, ql, qr, num_eqn, maxnx, num_ghost)

  implicit none

  integer,          intent(in)  :: num_eqn, maxnx, num_ghost
  double precision, intent(in)  :: q(num_eqn,maxnx+2*num_ghost)
  double precision, intent(out) :: ql(num_eqn,maxnx+2*num_ghost), qr(num_eqn,maxnx+2*num_ghost)

  integer :: i, m
  double precision :: {variables}

  do i = num_ghost, maxnx+num_ghost+1
    do m = 1, num_eqn
      {kernel}
    end do
  end do

end subroutine
'''

class FGenerator(pyweno.symbols.FGenerator):
  def __getitem__(self, idx):
    return pyweno.symbols.real('q(m,i{:+d})'.format(idx))

class FStarGenerator(pyweno.symbols.FStarGenerator):
  def __getitem__(self, idx):
    return pyweno.symbols.real([ 'ql(m,i)', 'qr(m,i)' ][idx])

pyweno.symbols.f  = FGenerator()
pyweno.symbols.fs = FStarGenerator()


#### open weno.f90 and write header

out = open('weno.f90', 'w')
out.write('''! This file was generated by: python weno.py
module weno
contains
''')


#### generate and write reconstruction functions for cell boundaries

for k in range(3, 3):

  print('generating reconstruction for k: %02d' % k)

  # smoothness
  beta = pyweno.symbolic.jiang_shu_smoothness_coefficients(k)

  # reconstructions: -1=left, 1=right
  varpi, split = pyweno.symbolic.optimal_weights(k, [ -1, 1 ])
  coeffs = pyweno.symbolic.reconstruction_coefficients(k, [ -1, 1 ])

  body = []
  body.append(kernel.smoothness(beta=beta))
  body.append(kernel.weights(varpi=varpi, split=split))
  body.append(kernel.reconstruction(coeffs=coeffs))

  variables = []
  variables.extend([ x['name'] for x in pyweno.symbols.sigma.all(k) ])
  variables.extend([ x['name'] for x in pyweno.symbols.omega.all(2, k, split) ])
  variables.extend([ x['name'] for x in pyweno.symbols.fr.all(2, k) ])

  # write function
  out.write(template.format(
      function='weno%d' % (2*k-1),
      variables=', &\n'.join(variables),
      kernel='\n'.join(body)))



"""
Tools for plotting data from gauges, gauge locations, etc.
"""

import os,sys,glob
import traceback
import warnings

import numpy as np

import clawpack.clawutil.data as clawdata

from clawpack.visclaw.frametools import set_show

plotter = 'matplotlib'
if plotter == 'matplotlib':
    if 'matplotlib' not in sys.modules:
        try:
            import matplotlib
            matplotlib.use('Agg')  # Use an image backend
        except:
            print("*** Error: problem importing matplotlib")

try:
    import pylab
except:
    print("*** Error: problem importing pylab")

# Gauge solution class
class GaugeSolution(object):
    r"""Object representing a single gauge"""

    def t1():
        doc = "(float) - Beginning of gauge recording time interval."
        def fget(self):
            if self.t is not None:
                return np.min(self.t)
            else:
                return None
        def fset(self, value):
            if self.q is None:
                if self.t is not None:
                    self.t[0] = value
                else:
                    self.t = [-np.infty,value]
            else:
                raise ValueError('Time interval already set.')
        return locals()
    t1 = property(**t1())

    def t2():
        doc = "(float) - End of gauge recording time interval."
        def fget(self):
            if self.t is not None:
                return np.max(self.t)
            else:
                return None
        def fset(self, value):
            if self.q is None:
                if self.t is not None:
                    self.t[1] = value
                else:
                    self.t = [-np.infty,value]
            else:
                raise ValueError('Time interval already set.')
        return locals()
    t2 = property(**t2())

    def location():
        doc = "(tuple) - Location of this gauge."
        def fget(self):
            if self._location is None:
                return ("Unknown","Unknown")
            return self._location
        def fset(self, value):
            if isinstance(value,tuple) or isinstance(value,list):
                self._location = value
            else:
                raise ValueError("Location information must be a list or tuple.")
        return locals()
    location = property(**location())

    def __init__(self, number, location=None):
        
        warnings.warn("This version of GaugeSolution is deprecated, use the ",
                      "class definition in clawpack.pyclaw.gauges instead.")

        # Gauge descriptors
        self.number = number
        self._location = None
        if location is not None:
            self.location = location

        # Data written out (usually)
        self.level = None
        self.t = None
        self.q = None

    def __repr__(self):
        # Make sure all necessary data has been set
        if self._location is None or self.t1 is None or self.t2 is None:
            output = None
        else:
            output = "%4i" % self.number
            for j in range(len(self.location)):
                output = " ".join((output,"%17.10e" % self.location[j]))
            output = " ".join((output,"%13.6e" % self.t1))
            output = " ".join((output,"%13.6e\n" % self.t2))
        return output

    def __str__(self):
        return ("Gauge %s: location = %s, t = [%s,%s]" % 
                                    (self.number,self.location,self.t1,self.t2))


#==========================================
def plotgauge(gaugeno, plotdata, verbose=False):
#==========================================

    """
    Plot all requested plots for a single gauge from the computation.
    The plots are requested by setting attributes of plotdata
    to ClawPlotFigure objects with plot_type="each_gauge".

    """

    if verbose:  
        gaugesoln = plotdata.getgauge(gaugeno)
        print('    Plotting gauge %s  at x = %s, y = %s ... '  \
                 % (gaugeno, gaugesoln.location[0], gaugesoln.location[1]))

    if plotdata.mode() == 'iplotclaw':
        pylab.ion()

        
    try:
        plotfigure_dict = plotdata.plotfigure_dict
    except:
        print('*** Error in plotgauge: plotdata missing plotfigure_dict')
        print('*** This should not happen')
        return None

    if len(plotfigure_dict) == 0:
        print('*** Warning in plotgauge: plotdata has empty plotfigure_dict')
        print('*** Apparently no figures to plot')




    # initialize current_data containing data that will be passed
    # to beforegauge, aftergauge, afteraxes commands
    current_data = clawdata.ClawData()
    current_data.add_attribute("user",{})   # for user specified attributes
                                           # to avoid potential conflicts
    current_data.add_attribute('plotdata',plotdata)
    current_data.add_attribute('gaugeno',gaugeno)

    # call beforegauge if present, which might define additional 
    # attributes in current_data or otherwise set up plotting for this
    # gauge.

    beforegauge =  getattr(plotdata, 'beforegauge', None)
    if beforegauge:
        if isinstance(beforegauge, str):
            # a string to be executed
            exec(beforegauge)
        else:
            # assume it's a function
            try:
                output = beforegauge(current_data)
                if output: current_data = output
            except:
                print('*** Error in beforegauge ***')
                raise



    # iterate over each single plot that makes up this gauge:
    # -------------------------------------------------------
 
    if plotdata._mode == 'iplotclaw':
        gaugesoln = plotdata.getgauge(gaugeno)
        gaugeloc = gaugesoln.location
        if len(gaugeloc) == 2:
            print('    Plotting Gauge %s  at x = %s, y = %s ... '  \
                     % (gaugeno, gaugeloc[0], gaugeloc[1]))
        elif len(gaugeloc) == 3:
            print('    Plotting Gauge %s  at x = %s, y = %s, z = %s ... '  \
                     % (gaugeno, gaugeloc[0], gaugeloc[1], gaugeloc[2]))
        requested_fignos = plotdata.iplotclaw_fignos
    else:
        requested_fignos = plotdata.print_fignos
    plotted_fignos = []

    plotdata = set_show(plotdata)   # set _show attributes for which figures
                                    # and axes should be shown.

    # loop over figures to appear for this gauge: 
    # -------------------------------------------

    for figname in plotdata._fignames:
        plotfigure = plotdata.plotfigure_dict[figname]
        if (not plotfigure._show) or (plotfigure.type != 'each_gauge'):
            continue  # skip to next figure 

        figno = plotfigure.figno
        if requested_fignos != 'all':
            if figno not in requested_fignos:
                continue # skip to next figure

        plotted_fignos.append(figno)


        if 'facecolor' not in plotfigure.kwargs:
            # Use white as default starting in v5.10.0
            # To use old default Clawpack tan, in setplot.py set:
            #     plotfigure.facecolor = \
            #           clawpack.visclaw.colormaps.clawpack_tan
            plotfigure.kwargs['facecolor'] = 'w'

        if plotfigure.figsize is not None:
            plotfigure.kwargs['figsize'] = plotfigure.figsize

        # create figure and set handle:
        plotfigure._handle = pylab.figure(num=figno, **plotfigure.kwargs)

        pylab.ioff()
        if plotfigure.clf_each_gauge:
            pylab.clf()

        try:
            plotaxes_dict = plotfigure.plotaxes_dict
        except:
            print('*** Error in plotgauge: plotdata missing plotaxes_dict')
            print('*** This should not happen')
            return  None

        if (len(plotaxes_dict) == 0) or (len(plotfigure._axesnames) == 0):
            print('*** Warning in plotgauge: plotdata has empty plotaxes_dict')
            print('*** Apparently no axes to plot in figno ',figno)

        # loop over axes to appear on this figure:
        # ----------------------------------------

        for axesname in plotfigure._axesnames:
            plotaxes = plotaxes_dict[axesname]
            if not plotaxes._show:
                continue   # skip this axes if no items show

            # create the axes:
            axescmd = getattr(plotaxes,'axescmd','subplot(1,1,1)')
            axescmd = 'plotaxes._handle = pylab.%s' % axescmd
            exec(axescmd)

            # loop over items:
            # ----------------

            for itemname in plotaxes._itemnames:
                
                plotitem = plotaxes.plotitem_dict[itemname]
                outdir = plotitem.outdir
                if outdir is None:
                    outdir = plotdata.outdir
                gaugesoln = plotdata.getgauge(gaugeno, outdir)

                current_data.add_attribute('gaugesoln',gaugesoln)
                current_data.add_attribute('q',gaugesoln.q)
                current_data.add_attribute('t',gaugesoln.t)

                if plotitem._show:
                    try:
                        output = plotgauge1(gaugesoln,plotitem,current_data)
                        if output: current_data = output
                        if verbose:  
                                print('      Plotted  plotitem ', itemname)
                    except:
                        print('*** Error in plotgauge: problem calling plotgauge1')
                        traceback.print_exc()
                        return None

            # end of loop over plotitems

    
            title_str = "%s at gauge %s" % (plotaxes.title,gaugeno)
            if plotaxes.title_fontsize is not None:
                plotaxes.title_kwargs['fontsize'] = plotaxes.title_fontsize
            pylab.title(title_str, **plotaxes.title_kwargs)
    
            if plotaxes.time_label is not None:
                if plotaxes.time_label_fontsize is not None:
                    plotaxes.time_label_kwargs['fontsize'] = \
                                    plotaxes.time_label_fontsize
                pylab.xlabel(plotaxes.time_label, **plotaxes.time_label_kwargs)
    
    
            # call an afteraxes function if present:
            afteraxes =  getattr(plotaxes, 'afteraxes', None)
            if afteraxes:
                if isinstance(afteraxes, str):
                    # a string to be executed
                    exec(afteraxes)
                else:
                    # assume it's a function
                    try:
                        current_data.add_attribute("plotaxes",plotaxes)
                        current_data.add_attribute("plotfigure",plotaxes._plotfigure)
                        output = afteraxes(current_data)
                        if output: current_data = output
                    except:
                        print('*** Error in afteraxes ***')
                        raise
    
            if plotaxes.scaled:
                pylab.axis('scaled')
    
            # set axes limits:
            if (plotaxes.xlimits is not None) & (type(plotaxes.xlimits) is not str):
                try:
                    pylab.xlim(plotaxes.xlimits[0], plotaxes.xlimits[1])
                except:
                    pass  # let axis be set automatically
            if (plotaxes.ylimits is not None) & (type(plotaxes.ylimits) is not str):
                try:
                    pylab.ylim(plotaxes.ylimits[0], plotaxes.ylimits[1])
                except:
                    pass  # let axis be set automatically
    
            if plotaxes.grid:
                pylab.grid(**plotaxes.grid_kwargs)
 
            if plotaxes.xticks_kwargs is not None:
                pylab.xticks(**plotaxes.xticks_kwargs)
            if plotaxes.yticks_kwargs is not None:
                pylab.yticks(**plotaxes.yticks_kwargs)

            if plotaxes.ylabel is not None:
                if plotaxes.ylabel_fontsize is not None:
                    plotaxes.ylabel_kwargs['fontsize'] = plotaxes.ylabel_fontsize
                pylab.ylabel(plotaxes.ylabel, **plotaxes.ylabel_kwargs)

            # end of loop over plotaxes
            
        # end of loop over plotfigures


    # call an aftergauge function if present:
    aftergauge =  getattr(plotdata, 'aftergauge', None)
    if aftergauge:
        if isinstance(aftergauge, str):
            # a string to be executed
            exec(aftergauge)
        else:
            # assume it's a function
            try:
                output = aftergauge(current_data)
                if output: current_data = output
            except:
                print('*** Error in aftergauge ***')
                raise


    if plotdata.mode() == 'iplotclaw':
        pylab.ion()
    for figno in plotted_fignos:
        pylab.figure(figno)
        pylab.draw()

    if verbose:
        print('    Done with plotgauge for gauge %i' % (gaugeno))

    
    # print the figure(s) to file(s) if requested:
    if (plotdata.mode() != 'iplotclaw') & plotdata.printfigs:
        # iterate over all figures that are to be printed:
        for figno in plotted_fignos:
            printfig(gaugeno=gaugeno, figno=figno, \
                    format=plotdata.print_format, plotdir=plotdata.plotdir,\
                    verbose=verbose)

    return current_data

    # end of plotgauge

    
#==================================================================
def plotgauge1(gaugesoln, plotitem, current_data):
#==================================================================
    """
    Make a 1d plot for a single plot item for the gauge solution in
    gaugesoln.

    The current_data object holds data that should be passed into
    afterpatch or afteraxes if these functions are defined.  The functions
    may add to this object, so this function should return the possibly
    modified current_data for use in other plotitems or in afteraxes or
    afterframe.

    """

    if not gaugesoln.is_valid():
        import warnings
        warnings.warn("Gauge has not been initialized properly.")
        return

    plotdata = plotitem._plotdata
    plotfigure = plotitem._plotfigure
    plotaxes = plotitem._plotaxes


    # the following plot parameters may be set, depending on what
    # plot_type was requested:

    plot_params = """
             plot_var  afterpatch  plotstyle color kwargs 
             plot_var2 fill_where map_2d_to_1d 
             """.split()

    # No amr_ parameters for gauge data.

    plot_var = plotitem.plot_var
    plot_type = plotitem.plot_type
    kwargs = plotitem.kwargs
    color = plotitem.color
    plotstyle = plotitem.plotstyle

    t = gaugesoln.t * plotaxes.time_scale

    if type(plot_var) is int:
        var = gaugesoln.q[plot_var,:]
    else:
        try:
            var = plot_var(current_data)
        except:
            print('Applying plot_var to current_data failed, try gaugesoln')
            try:
                var = plot_var(gaugesoln)
            except:
                raise Exception("Problem applying plot_var to gaugesoln")
    tmax = t.max()
    varmax = var.max()


    if (plot_type in ['1d_plot']) and (plotstyle != ''):
        if color:
            kwargs['color'] = color

        pobj = pylab.plot(t,var,plotstyle,**kwargs)


    elif plot_type == '1d_empty':
        # no plot to create (user might make one in afteritem or
        # afteraxes)
        pass

    else:
        raise ValueError("Unrecognized plot_type: %s" % plot_type)
        return None

    return current_data
    
def read_setgauges(outdir):
    """
    Read the info from gauges.data.
    """

    try:
        import clawpack.amrclaw.data as amrclaw
    except ImportError as e:
        print("You must have AMRClaw installed to plot gauges.")
        print("continuing...")
        return None

    setgauges = amrclaw.GaugeData()
    try:
        setgauges.read(outdir)
    except IOError as e:
        # No gauges.data file was found, ignore this exception
        pass

    return setgauges


def plot_gauge_locations(plotdata, gaugenos='all', mapc2p=None, \
                format_string='ko',add_labels=True, \
                markersize=5, fontsize=15, xoffset=0, yoffset=0):
    """
    Plot gauge locations on current axes.
    format_string specifies the symbol to be plotted.
    If add_labels==True then labels will also be added (gauge number).
    This routine determines locations from the file gauges.data in
    directory plotdata.rundir.  It does not require reading in the fort.gauge file
    produced by running the code.
    """

    from pylab import figure, plot, clf, title, text

    datadir = plotdata.rundir  # this should contain gauges.data

    try:
        setgauges = read_setgauges(plotdata.outdir)
    except:
        return

    if len(setgauges.gauges) == 0:
        print("*** plot_gauge_locations: No gauges specified in gauges.data")
        return

    if gaugenos=='all':
        gaugenos = setgauges.gauge_numbers
        gauges = setgauges.gauges
    else:
        gauges = []
        for gauge_num in gaugenos:
            for gauge in setgauges.gauges:
                if gauge_num == gauge[0]:
                    gauges.append(gauge)

    for gauge in gauges:
        try:
            xn,yn = gauge[1:3]
            if mapc2p:
                xn,yn = mapc2p(xn,yn)
            plot([xn], [yn], format_string, markersize=markersize)
            if add_labels: 
                xn = xn + xoffset
                yn = yn + yoffset
                text(xn,yn,'  %s' % gauge[0], fontsize=fontsize)
        except:
            print("*** plot_gauge_locations: warning: did not find x,y data for gauge ",gauge[0])


#------------------------------------------------------------------------
def printfig(fname='',gaugeno='', figno='', format='png', plotdir='.', \
             verbose=True):
#------------------------------------------------------------------------
    """
    Save the current plot to file fname or standard name from gauge/fig.
.  
    If fname is nonempty it is used as the filename, with extension
    determined by format if it does not already have a valid extension.

    If fname=='' then save to file gauge000NfigJ.ext  where N is the gauge
    number gaugeno passed in, J is the figure number figno passed in,
    and the extension ext is determined by format.  
    If figno='' then the figJ part is omitted.
    """

    if fname == '':
        fname = 'gauge' + str(gaugeno).rjust(4,'0') 
        if isinstance(figno,int):
            fname = fname + 'fig%s' % figno
    splitfname = os.path.splitext(fname)
    if splitfname[1] not in ('.png','.emf','.eps','.pdf'):
        fname = splitfname[0] + '.%s' % format
    if figno=='':
        figno = 1
    pylab.figure(figno)
    if plotdir != '.':
       fname = os.path.join(plotdir,fname)
    if verbose:  print('    Saving plot to file ', fname)
    pylab.savefig(fname)


#======================================================================
def printgauges(plotdata=None, verbose=True):
#======================================================================

    """
    Produce a set of png files for all the figures specified by plotdata.
    Also produce a set of html files for viewing the figures and navigating
    between them.  These will all be in directorey plotdata.plotdir.

    The ClawPlotData object plotdata will be initialized by a call to
    function setplot unless plotdata.setplot=False.  

    If plotdata.setplot=True then it is assumed that the current directory
    contains a module setplot.py that defines this function.

    If plotdata.setplot is a string then it is assumed this is the name of
    a module to import that contains the function setplot.

    If plotdata.setplot is a function then this function will be used.
    """

    import glob
    from clawpack.visclaw.data import ClawPlotData

    from clawpack.visclaw import plotpages


    if 'matplotlib' not in sys.modules:
        print('*** Error: matplotlib not found, no plots will be done')
        return plotdata
        
    if not isinstance(plotdata,ClawPlotData):
        print('*** Error, plotdata must be an object of type ClawPlotData')
        return plotdata

    plotdata._mode = 'printframes'

    plotdata = call_setplot(plotdata.setplot, plotdata)

    try:
        plotdata.rundir = os.path.abspath(plotdata.rundir)
        plotdata.outdir = os.path.abspath(plotdata.outdir)
        plotdata.plotdir = os.path.abspath(plotdata.plotdir)

        framenos = plotdata.print_framenos # frames to plot
        framenos = plotdata.print_framenos # frames to plot
        fignos = plotdata.print_fignos     # figures to plot at each frame
        fignames = {}                      # names to use in html files

        rundir = plotdata.rundir       # directory containing *.data files
        outdir = plotdata.outdir       # directory containing fort.* files
        plotdir = plotdata.plotdir     # where to put png and html files
        overwrite = plotdata.overwrite # ok to overwrite?
        msgfile = plotdata.msgfile     # where to write error messages
    except:
        print('*** Error in printframes: plotdata missing attribute')
        print('  *** plotdata = ',plotdata)
        return plotdata

    if fignos == 'all':
        fignos = plotdata._fignos
        #for (figname,plotfigure) in plotdata.plotfigure_dict.iteritems():
        #    fignos.append(plotfigure.figno)


    # filter out the fignos that will be empty, i.e.  plotfigure._show=False.
    plotdata = set_show(plotdata)
    fignos_to_show = []
    for figname in plotdata._fignames:
        figno = plotdata.plotfigure_dict[figname].figno
        if (figno in fignos) and plotdata.plotfigure_dict[figname]._show:
            fignos_to_show.append(figno)
    fignos = fignos_to_show
        
    # figure out what type each figure is:
    fignos_each_frame = []
    fignos_each_gauge = []
    fignos_each_run = []
    for figno in fignos:
        figname = plotdata._figname_from_num[figno]
        if plotdata.plotfigure_dict[figname].type == 'each_frame':
            fignos_each_frame.append(figno)
        if plotdata.plotfigure_dict[figname].type == 'each_gauge':
            fignos_each_gauge.append(figno)
        if plotdata.plotfigure_dict[figname].type == 'each_run':
            fignos_each_run.append(figno)
        

    rootdir = os.getcwd()

    # annoying fix needed when EPD is used for plotting under cygwin:
    if rootdir[0:9] == r'C:\cygwin' and outdir[0:9] != r'C:\cygwin':
        outdir = r'C:\cygwin' + outdir
        plotdata.outdir = outdir
    if rootdir[0:9] == r'C:\cygwin' and rundir[0:9] != r'C:\cygwin':
        rundir = r'C:\cygwin' + rundir
        plotdata.rundir = rundir
    if rootdir[0:9] == r'C:\cygwin' and plotdir[0:9] != r'C:\cygwin':
        plotdir = r'C:\cygwin' + plotdir
        plotdata.plotdir = plotdir

    try:
        os.chdir(rundir)
    except:
        print('*** Error: cannot move to run directory ',rundir)
        print('rootdir = ',rootdir)
        return plotdata


    if msgfile != '':
        sys.stdout = open(msgfile, 'w')
        sys.stderr = sys.stdout


    try:
        plotpages.cd_plotdir(plotdata)
    except:
        print("*** Error, aborting plotframes")
        return plotdata


    framefiles = glob.glob(os.path.join(plotdir,'frame*.png')) + \
                    glob.glob(os.path.join(plotdir,'frame*.html'))
    if overwrite:
        # remove any old versions:
        for file in framefiles:
            os.remove(file)
    else:
        if len(framefiles) > 1:
            print("*** Remove frame*.png and frame*.html and try again,")
            print("  or use overwrite=True in call to printframes")
            return plotdata

    
    # Create each of the figures
    #---------------------------

    try:
        os.chdir(outdir)
    except:
        print('*** Error printframes: cannot move to outdir = ',outdir)
        return plotdata


    fortfile = {}
    pngfile = {}
    frametimes = {}

    import glob
    for file in glob.glob('fort.q*'):
        frameno = int(file[7:10])
        fortfile[frameno] = file
        for figno in fignos_each_frame:
            pngfile[frameno,figno] = 'frame' + file[-4:] + 'fig%s.png' % figno
    
    if len(fortfile) == 0:
        print('*** No fort.q files found in directory ', os.getcwd())
        return plotdata
    
    # Discard frames that are not from latest run, based on
    # file modification time:
    framenos = only_most_recent(framenos, plotdata.outdir)

    numframes = len(framenos)

    print("Will plot %i frames numbered:" % numframes, framenos)
    print('Will make %i figure(s) for each frame, numbered: ' \
          % len(fignos_each_frame), fignos_each_frame)

    #fignames = {}
    #for figname in plotdata._fignames:
        #figno = plotdata.plotfigure_dict[figname].figno
        #fignames[figno] = figname
    # use new attribute:
    fignames = plotdata._figname_from_num


    for frameno in framenos:
        frametimes[frameno] = plotdata.getframe(frameno, plotdata.outdir).t

    plotdata.timeframes_framenos = framenos
    plotdata.timeframes_frametimes = frametimes
    plotdata.timeframes_fignos = fignos_each_frame
    plotdata.timeframes_fignames = fignames

    # Gauges:
    gaugenos = plotdata.print_gaugenos

    # Make html files for time frame figures:
    # ---------------------------------------

    os.chdir(plotdir)

    if plotdata.html:
        plotpages.timeframes2html(plotdata)
    
    # Make png files for all frames and gauges:
    # -----------------------------------------

    if not plotdata.printfigs:
        print("Using previously printed figure files")
    else:
        print("Now making png files for all figures...")
        for frameno in framenos:
            plotframe(frameno, plotdata, verbose)
            print('Frame %i at time t = %s' % (frameno, frametimes[frameno]))
        for gaugeno in gaugenos:
            plotgauge(gaugeno, plotdata, verbose)
            print('Gauge %i ' % gaugeno)


    if plotdata.latex:
        plotpages.timeframes2latex(plotdata)
    

    # Movie:
    #-------
    
    if plotdata.gif_movie:
        print('Making gif movies.  This may take some time....')
        for figno in fignos_each_frame:
            try:
                os.system('convert -delay 20 frame*fig%s.png moviefig%s.gif' \
                   % (figno,figno))
                print('    Created moviefig%s.gif' % figno)
            except:
                print('*** Error creating moviefig%s.gif' % figno)
    
    os.chdir(rootdir)

    # print out pointers to html index page:
    path_to_html_index = os.path.join(os.path.abspath(plotdata.plotdir), \
                               plotdata.html_index_fname)
    plotpages.print_html_pointers(path_to_html_index)

    # reset stdout for future print statements
    sys.stdout = sys.__stdout__

    return plotdata
    # end of printframes


def compare_gauges(outdir1, outdir2, gaugenos='all', q_components='all',
                    tol=0., verbose=True, plot=False):

    """
    Compare gauge output in two output directories.

    :Input:
     - *outdir1, outdir2* -- output directories
     - *gaugenos* -- list of gauge numbers to compare, or 'all' in which case
       outdir1/gauges.data will be used to determine gauge numbers.
     - *q_components* -- list of components of q to compare.
     - *tol* -- tolerance for checking equality
     - *verbose* -- print out dt and dq for each comparison?
     - *plot* -- if True, will produce a plot for each gauge, with a
       subfigure for each component of q.


    Returns True if everything matches to tolerance *tol*, else False.
    """

    from clawpack.visclaw.data import ClawPlotData
    from matplotlib import pyplot as plt
    
    if gaugenos == 'all':
        # attempt to read from gauges.data:
        try:
            setgauges1 = read_setgauges(outdir1)
            setgauges2 = read_setgauges(outdir2)
        except:
            print('*** could not read gauges.data from one of the outdirs')
            return
        gaugenos = setgauges1.gauge_numbers
        if setgauges2.gauge_numbers != gaugenos:
            print('*** warning -- outdirs have different sets of gauges')

        if len(gaugenos)==0:
            print("*** No gauges found in gauges.data")
            return

    plotdata1 = ClawPlotData()
    plotdata1.outdir = outdir1
    plotdata2 = ClawPlotData()
    plotdata2.outdir = outdir2

    matches = True
    for gaugeno in gaugenos:
        g1 = plotdata1.getgauge(gaugeno,verbose=verbose)
        t1 = g1.t
        q1 = g1.q

        g2 = plotdata2.getgauge(gaugeno,verbose=verbose)
        t2 = g2.t
        q2 = g2.q

        dt = abs(t1-t2).max()
        if verbose:
            print("Max difference in t[:] at gauge %s is %g" % (gaugeno,dt))
        matches = matches and (dt <= tol)

        if q_components == 'all':
            q_components = list(range(q1.shape[0]))

        for m in q_components:
            dq = abs(q1[m,:]-q2[m,:]).max()
            if verbose:
                print("Max difference in q[%s] at gauge %s is %g" % (m,gaugeno,dq))
            matches = matches and (dq <= tol)

        if plot:
            plt.figure(gaugeno)
            plt.clf()
            mq = len(q_components)
            for k,m in enumerate(q_components):
                plt.subplot(mq,1,k+1)
                plt.plot(g1.t,g1.q[m,:],'b',label='outdir1')
                plt.plot(g2.t,g2.q[m,:],'r',label='outdir2')
                plt.legend()
                plt.title('q[%s] at gauge number %s' % (m,gaugeno))
        
    return matches       
